# -*- coding: utf-8 -*-

"""Tests Utilities

Refactor utility functions and classes
"""

DictProxyType = type(object.__dict__)


def dprint(msg):
    # Debugging helper to trace thread-related tests.
    if 0:
        print(msg)


def is_clr_module(ob):
    return type(ob).__name__ == 'ModuleObject'


def is_clr_root_module(ob):
    # in Python 3 the clr module is a normal python module
    return ob.__name__ == "clr"


def is_clr_class(ob):
    return type(ob).__name__ == 'CLR Metatype'  # for now


class ClassicClass:
    def kind(self):
        return "classic"


class NewStyleClass(object):
    def kind(self):
        return "new-style"


class GenericHandler(object):
    """A generic handler to test event callbacks."""

    def __init__(self):
        self.value = None

    def handler(self, sender, args):
        self.value = args.value


class VariableArgsHandler(object):
    """A variable args handler to test event callbacks."""

    def __init__(self):
        self.value = None

    def handler(self, *args):
        ob, eventargs = args
        self.value = eventargs.value


class CallableHandler(object):
    """A callable handler to test event callbacks."""

    def __init__(self):
        self.value = None

    def __call__(self, sender, args):
        self.value = args.value


class VarCallableHandler(object):
    """A variable args callable handler to test event callbacks."""

    def __init__(self):
        self.value = None

    def __call__(self, *args):
        ob, eventargs = args
        self.value = eventargs.value


class StaticMethodHandler(object):
    """A static method handler to test event callbacks."""

    value = None

    def handler(sender, args):
        StaticMethodHandler.value = args.value

    handler = staticmethod(handler)


class ClassMethodHandler(object):
    """A class method handler to test event callbacks."""

    value = None

    def handler(cls, sender, args):
        cls.value = args.value

    handler = classmethod(handler)


class MultipleHandler(object):
    """A generic handler to test multiple callbacks."""

    def __init__(self):
        self.value = 0

    def handler(self, sender, args):
        self.value += args.value

    def count(self):
        self.value += 1
        return 'ok'


class HelloClass(object):
    def hello(self):
        return "hello"

    def __call__(self):
        return "hello"

    @staticmethod
    def s_hello():
        return "hello"

    @classmethod
    def c_hello(cls):
        return "hello"


def hello_func():
    return "hello"



# -*- coding: utf-8 -*-

"""Test exception support."""

import sys

import System
import pytest
import pickle


def test_unified_exception_semantics():
    """Test unified exception semantics."""
    e = System.Exception('Something bad happened')
    assert isinstance(e, Exception)
    assert isinstance(e, System.Exception)


def test_standard_exception_attributes():
    """Test accessing standard exception attributes."""
    from System import OverflowException
    from Python.Test import ExceptionTest

    e = ExceptionTest.GetExplicitException()
    assert isinstance(e, OverflowException)

    assert e.Message == 'error'

    e.Source = 'Test Suite'
    assert e.Source == 'Test Suite'

    v = e.ToString()
    assert len(v) > 0


def test_extended_exception_attributes():
    """Test accessing extended exception attributes."""
    from Python.Test import ExceptionTest, ExtendedException
    from System import OverflowException

    e = ExceptionTest.GetExtendedException()
    assert isinstance(e, ExtendedException)
    assert isinstance(e, OverflowException)
    assert isinstance(e, System.Exception)

    assert e.Message == 'error'

    e.Source = 'Test Suite'
    assert e.Source == 'Test Suite'

    v = e.ToString()
    assert len(v) > 0

    assert e.ExtraProperty == 'extra'
    e.ExtraProperty = 'changed'
    assert e.ExtraProperty == 'changed'

    assert e.GetExtraInfo() == 'changed'


def test_raise_class_exception():
    """Test class exception propagation."""
    from System import NullReferenceException

    with pytest.raises(NullReferenceException) as cm:
        raise NullReferenceException

    exc = cm.value
    assert isinstance(exc, NullReferenceException)


def test_exc_info():
    """Test class exception propagation.
    Behavior of exc_info changed in Py3. Refactoring its test"""
    from System import NullReferenceException
    try:
        raise NullReferenceException("message")
    except Exception as exc:
        type_, value, tb = sys.exc_info()
        assert type_ is NullReferenceException
        assert value.Message == "message"
        assert exc.Message == "message"
        # FIXME: Lower-case message isn't implemented
        # self.assertTrue(exc.message == "message")
        assert value is exc


def test_raise_class_exception_with_value():
    """Test class exception propagation with associated value."""
    from System import NullReferenceException

    with pytest.raises(NullReferenceException) as cm:
        raise NullReferenceException('Aiiieee!')

    exc = cm.value
    assert isinstance(exc, NullReferenceException)
    assert exc.Message == 'Aiiieee!'


def test_raise_instance_exception():
    """Test instance exception propagation."""
    from System import NullReferenceException

    with pytest.raises(NullReferenceException) as cm:
        raise NullReferenceException()

    exc = cm.value
    assert isinstance(exc, NullReferenceException)
    assert len(exc.Message) > 0


def test_raise_instance_exception_with_args():
    """Test instance exception propagation with args."""
    from System import NullReferenceException

    with pytest.raises(NullReferenceException) as cm:
        raise NullReferenceException("Aiiieee!")

    exc = cm.value
    assert isinstance(exc, NullReferenceException)
    assert exc.Message == 'Aiiieee!'


def test_managed_exception_propagation():
    """Test propagation of exceptions raised in managed code."""
    from System import Decimal, OverflowException

    with pytest.raises(OverflowException):
        Decimal.ToInt64(Decimal.MaxValue)


def test_managed_exception_conversion():
    """Test conversion of managed exceptions."""
    from System import OverflowException
    from Python.Test import ExceptionTest

    e = ExceptionTest.GetBaseException()
    assert isinstance(e, System.Exception)

    e = ExceptionTest.GetExplicitException()
    assert isinstance(e, OverflowException)
    assert isinstance(e, System.Exception)

    e = ExceptionTest.GetWidenedException()
    assert isinstance(e, OverflowException)
    assert isinstance(e, System.Exception)

    v = ExceptionTest.SetBaseException(System.Exception('error'))
    assert v

    v = ExceptionTest.SetExplicitException(OverflowException('error'))
    assert v

    v = ExceptionTest.SetWidenedException(OverflowException('error'))
    assert v


def test_catch_exception_from_managed_method():
    """Test catching an exception from a managed method."""
    from Python.Test import ExceptionTest
    from System import OverflowException

    with pytest.raises(OverflowException) as cm:
        ExceptionTest().ThrowException()

    e = cm.value
    assert isinstance(e, OverflowException)


def test_catch_exception_from_managed_property():
    """Test catching an exception from a managed property."""
    from Python.Test import ExceptionTest
    from System import OverflowException

    with pytest.raises(OverflowException) as cm:
        _ = ExceptionTest().ThrowProperty

    e = cm.value
    assert isinstance(e, OverflowException)

    with pytest.raises(OverflowException) as cm:
        ExceptionTest().ThrowProperty = 1

    e = cm.value
    assert isinstance(e, OverflowException)


def test_catch_exception_managed_class():
    """Test catching the managed class of an exception."""
    from System import OverflowException

    with pytest.raises(OverflowException):
        raise OverflowException('overflow')


def test_catch_exception_python_class():
    """Test catching the python class of an exception."""
    from System import OverflowException

    with pytest.raises(Exception):
        raise OverflowException('overflow')


def test_catch_exception_base_class():
    """Test catching the base of an exception."""
    from System import OverflowException, ArithmeticException

    with pytest.raises(ArithmeticException):
        raise OverflowException('overflow')


def test_catch_exception_nested_base_class():
    """Test catching the nested base of an exception."""
    from System import OverflowException, SystemException

    with pytest.raises(SystemException):
        raise OverflowException('overflow')


def test_catch_exception_with_assignment():
    """Test catching an exception with assignment."""
    from System import OverflowException

    with pytest.raises(OverflowException) as cm:
        raise OverflowException('overflow')

    e = cm.value
    assert isinstance(e, OverflowException)


def test_catch_exception_unqualified():
    """Test catching an unqualified exception."""
    from System import OverflowException

    try:
        raise OverflowException('overflow')
    except:
        pass
    else:
        self.fail("failed to catch unqualified exception")


def test_catch_baseexception():
    """Test catching an unqualified exception with BaseException."""
    from System import OverflowException

    with pytest.raises(BaseException):
        raise OverflowException('overflow')


def test_apparent_module_of_exception():
    """Test the apparent module of an exception."""
    from System import OverflowException

    assert System.Exception.__module__ == 'System'
    assert OverflowException.__module__ == 'System'


def test_str_of_exception():
    """Test the str() representation of an exception."""
    from System import NullReferenceException, Convert, FormatException

    e = NullReferenceException('')
    assert str(e) == ''

    e = NullReferenceException('Something bad happened')
    assert str(e).startswith('Something bad happened')

    with pytest.raises(FormatException) as cm:
        Convert.ToDateTime('this will fail')


def test_python_compat_of_managed_exceptions():
    """Test managed exceptions compatible with Python's implementation"""
    from System import OverflowException
    msg = "Simple message"

    e = OverflowException(msg)
    assert str(e) == msg

    assert e.args == (msg,)
    assert isinstance(e.args, tuple)
    strexp = "OverflowException('Simple message"
    assert repr(e)[:len(strexp)] == strexp


def test_exception_is_instance_of_system_object():
    """Test behavior of isinstance(<managed exception>, System.Object)."""
    # This is an anti-test, in that this is a caveat of the current
    # implementation. Because exceptions are not allowed to be new-style
    # classes, we wrap managed exceptions in a general-purpose old-style
    # class that delegates to the wrapped object. This makes _almost_
    # everything work as expected, except that an isinstance check against
    # System.Object will fail for a managed exception (because a new
    # style class cannot appear in the __bases__ of an old-style class
    # without causing a crash in the CPython interpreter). This test is
    # here mainly to remind me to update the caveat in the documentation
    # one day when when exceptions can be new-style classes.

    # This behavior is now over-shadowed by the implementation of
    # __instancecheck__ (i.e., overloading isinstance), so for all Python
    # version >= 2.6 we expect isinstance(<managed exception>, Object) to
    # be true, even though it does not really subclass Object.
    from System import OverflowException, Object

    o = OverflowException('error')

    if sys.version_info >= (2, 6):
        assert isinstance(o, Object)
    else:
        assert not isinstance(o, Object)


def test_pickling_exceptions():
    exc = System.Exception("test")
    dumped = pickle.dumps(exc)
    loaded = pickle.loads(dumped)

    assert exc.args == loaded.args


def test_chained_exceptions():
    from Python.Test import ExceptionTest

    with pytest.raises(Exception) as cm:
        ExceptionTest.ThrowChainedExceptions()

    exc = cm.value

    msgs = ("Outer exception",
            "Inner exception",
            "Innermost exception",)
    for msg in msgs:
        assert exc.Message == msg
        assert exc.__cause__ == exc.InnerException
        exc = exc.__cause__

def test_iteration_exception():
    from Python.Test import ExceptionTest
    from System import OverflowException

    exception = OverflowException("error")

    val = ExceptionTest.ThrowExceptionInIterator(exception).__iter__()
    assert next(val) == 1
    assert next(val) == 2
    with pytest.raises(OverflowException) as cm:
        next(val)

    exc = cm.value

    assert exc == exception

    # after exception is thrown iterator is no longer valid
    with pytest.raises(StopIteration):
        next(val)


def test_iteration_innerexception():
    from Python.Test import ExceptionTest
    from System import OverflowException

    exception = System.Exception("message", OverflowException("error"))

    val = ExceptionTest.ThrowExceptionInIterator(exception).__iter__()
    assert next(val) == 1
    assert next(val) == 2
    with pytest.raises(OverflowException) as cm:
        next(val)

    exc = cm.value

    assert exc == exception.InnerException

    # after exception is thrown iterator is no longer valid
    with pytest.raises(StopIteration):
        next(val)



# -*- coding: utf-8 -*-
# TODO: move tests one out of src to project root.
# TODO: travis has numpy on their workers. Maybe add tests?

"""Helpers for testing."""

import ctypes
import os
import sys
import sysconfig
from subprocess import check_call
from tempfile import mkdtemp
import shutil

import pytest

from pythonnet import set_runtime

# Add path for `Python.Test`
cwd = os.path.dirname(__file__)
fixtures_path = os.path.join(cwd, "fixtures")
sys.path.append(fixtures_path)

def pytest_addoption(parser):
    parser.addoption(
        "--runtime",
        action="store",
        default="default",
        help="Must be one of default, netcore, netfx and mono"
    )

collect_ignore = []

def pytest_configure(config):
    global bin_path
    if "clr" in sys.modules:
        # Already loaded (e.g. by the C# test runner), skip build
        import clr
        clr.AddReference("Python.Test")
        return

    runtime_opt = config.getoption("runtime")

    test_proj_path = os.path.join(cwd, "..", "src", "testing")

    if runtime_opt not in ["netcore", "netfx", "mono", "default"]:
        raise RuntimeError(f"Invalid runtime: {runtime_opt}")

    bin_path = mkdtemp()

    # tmpdir_factory.mktemp(f"pythonnet-{runtime_opt}")

    fw = "net6.0" if runtime_opt == "netcore" else "netstandard2.0"

    check_call(["dotnet", "publish", "-f", fw, "-o", bin_path, test_proj_path])

    sys.path.append(bin_path)

    if runtime_opt == "default":
        pass
    elif runtime_opt == "netfx":
        from clr_loader import get_netfx
        runtime = get_netfx()
        set_runtime(runtime)
    elif runtime_opt == "mono":
        from clr_loader import get_mono
        runtime = get_mono()
        set_runtime(runtime)
    elif runtime_opt == "netcore":
        from clr_loader import get_coreclr
        rt_config_path = os.path.join(bin_path, "Python.Test.runtimeconfig.json")
        runtime = get_coreclr(rt_config_path)
        set_runtime(runtime)

    import clr
    clr.AddReference("Python.Test")

    soft_mode = False
    try:
        os.environ['PYTHONNET_SHUTDOWN_MODE'] == 'Soft'
    except: pass

    if config.getoption("--runtime") == "netcore" or soft_mode\
        :
        collect_ignore.append("domain_tests/test_domain_reload.py")
    else:
        domain_tests_dir = os.path.join(os.path.dirname(__file__), "domain_tests")
        bin_path = os.path.join(domain_tests_dir, "bin")
        check_call(["dotnet", "build", domain_tests_dir, "-o", bin_path])




def pytest_unconfigure(config):
    global bin_path
    try:
        shutil.rmtree(bin_path)
    except Exception:
        pass

def pytest_report_header(config):
    """Generate extra report headers"""
    # FIXME: https://github.com/pytest-dev/pytest/issues/2257
    is_64bits = sys.maxsize > 2**32
    arch = "x64" if is_64bits else "x86"
    ucs = ctypes.sizeof(ctypes.c_wchar)
    libdir = sysconfig.get_config_var("LIBDIR")
    shared = bool(sysconfig.get_config_var("Py_ENABLE_SHARED"))

    header = ("Arch: {arch}, UCS: {ucs}, LIBDIR: {libdir}, "
              "Py_ENABLE_SHARED: {shared}".format(**locals()))
    return header


@pytest.fixture()
def filepath():
    """Returns full filepath for file in `fixtures` directory."""

    def make_filepath(filename):
        # http://stackoverflow.com/questions/18011902/parameter-to-a-fixture
        return os.path.join(fixtures_path, filename)

    return make_filepath



# -*- coding: utf-8 -*-

"""Test CLR modules and the CLR import hook."""

import clr
import time
import types
import warnings
from fnmatch import fnmatch

import pytest

from .utils import is_clr_class, is_clr_module, is_clr_root_module


# testImplicitAssemblyLoad() passes on deprecation warning; perfect! #
# clr.AddReference('System.Windows.Forms')

def test_import_hook_works():
    """Test that the import hook works correctly both using the
       included runtime and an external runtime. This must be
       the first test run in the unit tests!"""
    from System import String


def test_import_clr():
    import clr
    assert is_clr_root_module(clr)


def test_version_clr():
    import clr
    assert clr.__version__ >= "3.0.0"


def test_preload_var():
    import clr
    assert clr.getPreload() is False, clr.getPreload()
    clr.setPreload(False)
    assert clr.getPreload() is False, clr.getPreload()
    try:
        clr.setPreload(True)
        assert clr.getPreload() is True, clr.getPreload()

        import System.Configuration
        content = dir(System.Configuration)
        assert len(content) > 10, content
    finally:
        clr.setPreload(False)


def test_module_interface():
    """Test the interface exposed by CLR module objects."""
    import System
    assert type(System.__dict__) == type({})
    assert System.__name__ == 'System'
    # the filename can be any module from the System namespace
    # (eg System.Data.dll or System.dll, but also mscorlib.dll)
    system_file = System.__file__
    assert fnmatch(system_file, "*System*.dll") or fnmatch(system_file, "*mscorlib.dll"), \
        "unexpected System.__file__: " + system_file
    assert System.__doc__.startswith("Namespace containing types from the following assemblies:")
    assert is_clr_class(System.String)
    assert is_clr_class(System.Int32)


def test_simple_import():
    """Test simple import."""
    import System
    assert is_clr_module(System)
    assert System.__name__ == 'System'

    import sys
    assert isinstance(sys, types.ModuleType)
    assert sys.__name__ == 'sys'

    import http.client as httplib
    assert isinstance(httplib, types.ModuleType)
    assert httplib.__name__ == 'http.client'


def test_simple_import_with_alias():
    """Test simple import with aliasing."""
    import System as mySystem
    assert is_clr_module(mySystem)
    assert mySystem.__name__ == 'System'

    import sys as mySys
    assert isinstance(mySys, types.ModuleType)
    assert mySys.__name__ == 'sys'

    import http.client as myHttplib
    assert isinstance(myHttplib, types.ModuleType)
    assert myHttplib.__name__ == 'http.client'


def test_dotted_name_import():
    """Test dotted-name import."""
    import System.Reflection
    assert is_clr_module(System.Reflection)
    assert System.Reflection.__name__ == 'System.Reflection'

    import xml.dom
    assert isinstance(xml.dom, types.ModuleType)
    assert xml.dom.__name__ == 'xml.dom'


def test_multiple_dotted_name_import():
    """Test an import bug with multiple dotted imports."""

    import System.Reflection
    assert is_clr_module(System.Reflection)
    assert System.Reflection.__name__ == 'System.Reflection'
    import System.Reflection
    assert is_clr_module(System.Reflection)
    assert System.Reflection.__name__ == 'System.Reflection'


def test_dotted_name_import_with_alias():
    """Test dotted-name import with aliasing."""
    import System.Reflection as SysRef
    assert is_clr_module(SysRef)
    assert SysRef.__name__ == 'System.Reflection'

    import xml.dom as myDom
    assert isinstance(myDom, types.ModuleType)
    assert myDom.__name__ == 'xml.dom'


def test_simple_import_from():
    """Test simple 'import from'."""
    from System import Reflection
    assert is_clr_module(Reflection)
    assert Reflection.__name__ == 'System.Reflection'

    from xml import dom
    assert isinstance(dom, types.ModuleType)
    assert dom.__name__ == 'xml.dom'


def test_simple_import_from_with_alias():
    """Test simple 'import from' with aliasing."""
    from System import Collections as Coll
    assert is_clr_module(Coll)
    assert Coll.__name__ == 'System.Collections'

    from xml import dom as myDom
    assert isinstance(myDom, types.ModuleType)
    assert myDom.__name__ == 'xml.dom'


def test_dotted_name_import_from():
    """Test dotted-name 'import from'."""
    from System.Collections import Specialized
    assert is_clr_module(Specialized)
    assert Specialized.__name__ == 'System.Collections.Specialized'

    from System.Collections.Specialized import StringCollection
    assert is_clr_class(StringCollection)
    assert StringCollection.__name__ == 'StringCollection'

    from xml.dom import pulldom
    assert isinstance(pulldom, types.ModuleType)
    assert pulldom.__name__ == 'xml.dom.pulldom'

    from xml.dom.pulldom import PullDOM
    assert isinstance(PullDOM, type)
    assert PullDOM.__name__ == 'PullDOM'


def test_dotted_name_import_from_with_alias():
    """Test dotted-name 'import from' with aliasing."""
    from System.Collections import Specialized as Spec
    assert is_clr_module(Spec)
    assert Spec.__name__ == 'System.Collections.Specialized'

    from System.Collections.Specialized import StringCollection as SC
    assert is_clr_class(SC)
    assert SC.__name__ == 'StringCollection'

    from xml.dom import pulldom as myPulldom
    assert isinstance(myPulldom, types.ModuleType)
    assert myPulldom.__name__ == 'xml.dom.pulldom'

    from xml.dom.pulldom import PullDOM as myPullDOM
    assert isinstance(myPullDOM, type)
    assert myPullDOM.__name__ == 'PullDOM'


def test_from_module_import_star():
    """Test from module import * behavior."""
    clr.AddReference("System")

    count = len(locals().keys())
    m = __import__('System', globals(), locals(), ['*'])
    assert m.__name__ == 'System'
    assert is_clr_module(m)
    assert len(locals().keys()) > count + 1


def test_implicit_assembly_load():
    """Test implicit assembly loading via import."""
    with pytest.raises(ImportError):
        # MS.Build should not have been added as a reference yet
        # (and should exist for mono)

        # The implicit behavior has been disabled in 3.0
        # therefore this should fail
        import Microsoft.Build

    with warnings.catch_warnings(record=True) as w:
        try:
            clr.AddReference("System.Windows.Forms")
        except Exception:
            pytest.skip()

        import System.Windows.Forms as Forms
        assert is_clr_module(Forms)
        assert Forms.__name__ == 'System.Windows.Forms'
        from System.Windows.Forms import Form
        assert is_clr_class(Form)
        assert Form.__name__ == 'Form'
        assert len(w) == 0


def test_explicit_assembly_load():
    """Test explicit assembly loading using standard CLR tools."""
    from System.Reflection import Assembly
    import System, sys

    assembly = Assembly.LoadWithPartialName('Microsoft.CSharp')
    assert assembly is not None

    import Microsoft.CSharp
    assert 'Microsoft.CSharp' in sys.modules

    assembly = Assembly.LoadWithPartialName('SpamSpamSpamSpamEggsAndSpam')
    assert assembly is None


def test_implicit_load_already_valid_namespace():
    """Test implicit assembly load over an already valid namespace."""
    # In this case, the mscorlib assembly (loaded by default) defines
    # a number of types in the System namespace. There is also a System
    # assembly, which is _not_ loaded by default, which also contains
    # types in the System namespace. The desired behavior is for the
    # Python runtime to "do the right thing", allowing types from both
    # assemblies to be found in the System module implicitly.
    import System
    assert is_clr_class(System.UriBuilder)


def test_import_non_existant_module():
    """Test import failure for a non-existent module."""
    with pytest.raises(ImportError):
        import System.SpamSpamSpam


def test_lookup_no_namespace_type():
    """Test lookup of types without a qualified namespace."""
    import Python.Test
    import clr
    assert is_clr_class(clr.NoNamespaceType)


def test_module_lookup_recursion():
    """Test for recursive lookup handling."""

    with pytest.raises(ImportError):
        from System import System

    with pytest.raises(AttributeError):
        import System
        _ = System.System


def test_module_get_attr():
    """Test module getattr behavior."""

    import System
    import System.Runtime

    int_type = System.Int32
    assert is_clr_class(int_type)

    module = System.Runtime
    assert is_clr_module(module)

    with pytest.raises(AttributeError):
        _ = System.Spam

    with pytest.raises(TypeError):
        _ = getattr(System, 1)


def test_module_attr_abuse():
    """Test handling of attempts to set module attributes."""

    # It would be safer to use a dict-proxy as the __dict__ for CLR
    # modules, but as of Python 2.3 some parts of the CPython runtime
    # like dir() will fail if a module dict is not a real dictionary.

    def test():
        import System
        System.__dict__['foo'] = 0
        return 1

    assert test()


def test_module_type_abuse():
    """Test handling of attempts to break the module type."""
    import System
    mtype = type(System)

    with pytest.raises(TypeError):
        mtype.__getattribute__(0, 'spam')

    with pytest.raises(TypeError):
        mtype.__setattr__(0, 'spam', 1)

    with pytest.raises(TypeError):
        mtype.__repr__(0)


def test_clr_list_assemblies():
    from clr import ListAssemblies
    verbose = list(ListAssemblies(True))
    short = list(ListAssemblies(False))
    assert u'System' in short
    assert u'Culture=' in verbose[0]
    assert u'Version=' in verbose[0]


def test_clr_add_reference():
    from clr import AddReference
    from System.IO import FileNotFoundException
    for name in ("System", "Python.Runtime"):
        assy = AddReference(name)
        assy_name = assy.GetName().Name
        assert assy_name == name

    with pytest.raises(FileNotFoundException):
        AddReference("somethingtotallysilly")

def test_clr_get_clr_type():
    """Test clr.GetClrType()."""
    from clr import GetClrType
    import System
    from System import IComparable
    from System import ArgumentException
    assert GetClrType(System.String).FullName == "System.String"
    comparable = GetClrType(IComparable)
    assert comparable.FullName == "System.IComparable"
    assert comparable.IsInterface
    assert GetClrType(int).FullName == "Python.Runtime.PyInt"
    assert GetClrType(str).FullName == "System.String"
    assert GetClrType(float).FullName == "System.Double"
    dblarr = System.Array[System.Double]
    assert GetClrType(dblarr).FullName == "System.Double[]"

    with pytest.raises(TypeError):
        GetClrType(1)
    with pytest.raises(TypeError):
        GetClrType("thiswillfail")

def test_assembly_load_thread_safety():
    from Python.Test import ModuleTest
    # spin up .NET thread which loads assemblies and triggers AppDomain.AssemblyLoad event
    ModuleTest.RunThreads()
    time.sleep(1e-5)
    for _ in range(1, 100):
        # call import clr, which in AssemblyManager.GetNames iterates through the loaded types
        import clr
        # import some .NET types
        from System import DateTime
        from System import Guid
        from System.Collections.Generic import Dictionary
        _ = Dictionary[Guid, DateTime]()
    ModuleTest.JoinThreads()

@pytest.mark.skipif()
def test_assembly_load_recursion_bug():
    """Test fix for recursion bug documented in #627"""
    sys_config = pytest.importorskip(
        "System.Configuration", reason="System.Configuration can't be imported"
    )
    content = dir(sys_config.ConfigurationManager)
    assert len(content) > 5, content



"""Test sys.argv state."""

import sys
from subprocess import check_output
from ast import literal_eval


def test_sys_argv_state(filepath):
    """Test sys.argv state doesn't change after clr import.
    To better control the arguments being passed, test on a fresh python
    instance with specific arguments"""

    script = filepath("argv-fixture.py")
    out = check_output([sys.executable, script, "foo", "bar"])
    out = literal_eval(out.decode("ascii"))
    assert out[-2:] == ["foo", "bar"]



# -*- coding: utf-8 -*-
# TODO: Add tests for ClassicClass, NewStyleClass?

"""Test CLR class support."""

import clr
import Python.Test as Test
import System
import pytest

from .utils import DictProxyType


def test_basic_reference_type():
    """Test usage of CLR defined reference types."""
    assert System.String.Empty == ""

def test_basic_value_type():
    """Test usage of CLR defined value types."""
    assert System.Int32.MaxValue == 2147483647


def test_class_standard_attrs():
    """Test standard class attributes."""
    from Python.Test import ClassTest

    assert ClassTest.__name__ == 'ClassTest'
    assert ClassTest.__module__ == 'Python.Test'
    assert isinstance(ClassTest.__dict__, DictProxyType)
    assert len(ClassTest.__doc__) > 0

def test_class_docstrings():
    """Test standard class docstring generation"""
    from Python.Test import ClassTest

    value = 'Void .ctor()'
    assert ClassTest.__doc__ == value


def test_class_default_str():
    """Test the default __str__ implementation for managed objects."""
    s = System.String("this is a test")
    assert str(s) == "this is a test"


def test_class_default_repr():
    """Test the default __repr__ implementation for managed objects."""
    s = System.String("this is a test")
    assert repr(s).startswith("<System.String object")


def test_non_public_class():
    """Test that non-public classes are inaccessible."""
    with pytest.raises(ImportError):
        from Python.Test import InternalClass

    with pytest.raises(AttributeError):
        _ = Test.InternalClass

def test_non_exported():
    """Test [PyExport(false)]"""
    with pytest.raises(ImportError):
        from Python.Test import NonExportable

    with pytest.raises(AttributeError):
        _ = Test.NonExportable


def test_basic_subclass():
    """Test basic subclass of a managed class."""
    from System.Collections import Hashtable

    class MyTable(Hashtable):
        def how_many(self):
            return self.Count

    table = MyTable()

    assert table.__class__.__name__.endswith('MyTable')
    assert type(table).__name__.endswith('MyTable')
    assert len(table.__class__.__bases__) == 1
    assert table.__class__.__bases__[0] == Hashtable

    assert table.how_many() == 0
    assert table.Count == 0

    table.set_Item('one', 'one')

    assert table.how_many() == 1
    assert table.Count == 1


def test_subclass_with_no_arg_constructor():
    """Test subclass of a managed class with a no-arg constructor."""
    from Python.Test import ClassCtorTest1

    class SubClass(ClassCtorTest1):
        def __init__(self, name):
            self.name = name

    # This failed in earlier versions
    _ = SubClass('test')


def test_subclass_with_various_constructors():
    """Test subclass of a managed class with various constructors."""
    from Python.Test import ClassCtorTest2

    class SubClass(ClassCtorTest2):
        def __init__(self, v):
            ClassCtorTest2.__init__(self)
            self.value = v

    inst = SubClass('test')
    assert inst.value == 'test'

    class SubClass2(ClassCtorTest2):
        def __init__(self, v):
            ClassCtorTest2.__init__(self)
            self.value = v

    inst = SubClass2('test')
    assert inst.value == 'test'


def test_struct_construction():
    """Test construction of structs."""

    from Python.Test import Point

    p = Point()
    assert p.X == 0
    assert p.Y == 0

    p = Point(0, 0)
    assert p.X == 0
    assert p.Y == 0

    p.X = 10
    p.Y = 10

    assert p.X == 10
    assert p.Y == 10

    # test strange __new__ interactions

    # test weird metatype
    # test recursion
    # test


def test_ienumerable_iteration():
    """Test iteration over objects supporting IEnumerable."""
    from Python.Test import ClassTest

    list_ = ClassTest.GetArrayList()

    for item in list_:
        assert (item > -1) and (item < 10)

    dict_ = ClassTest.GetHashtable()

    for item in dict_:
        cname = item.__class__.__name__
        assert cname.endswith('DictionaryEntry')


def test_ienumerator_iteration():
    """Test iteration over objects supporting IEnumerator."""
    from Python.Test import ClassTest

    chars = ClassTest.GetEnumerator()

    for item in chars:
        assert item in 'test string'

def test_iterable():
    """Test what objects are Iterable"""
    from collections.abc import Iterable
    from Python.Test import ClassTest

    assert isinstance(System.String.Empty, Iterable)
    assert isinstance(ClassTest.GetArrayList(), Iterable)
    assert isinstance(ClassTest.GetEnumerator(), Iterable)
    assert (not isinstance(ClassTest, Iterable))
    assert (not isinstance(ClassTest(), Iterable))

    class ShouldBeIterable(ClassTest):
        def __iter__(self):
            return iter([])

    assert isinstance(ShouldBeIterable(), Iterable)


def test_override_get_item():
    """Test managed subclass overriding __getitem__."""
    from System.Collections import Hashtable

    class MyTable(Hashtable):
        def __getitem__(self, key):
            value = Hashtable.__getitem__(self, key)
            return 'my ' + str(value)

    table = MyTable()
    table['one'] = 'one'
    table['two'] = 'two'
    table['three'] = 'three'

    assert table['one'] == 'my one'
    assert table['two'] == 'my two'
    assert table['three'] == 'my three'

    assert table.Count == 3


def test_override_set_item():
    """Test managed subclass overriding __setitem__."""
    from System.Collections import Hashtable

    class MyTable(Hashtable):
        def __setitem__(self, key, value):
            value = 'my ' + str(value)
            Hashtable.__setitem__(self, key, value)

    table = MyTable()
    table['one'] = 'one'
    table['two'] = 'two'
    table['three'] = 'three'

    assert table['one'] == 'my one'
    assert table['two'] == 'my two'
    assert table['three'] == 'my three'

    assert table.Count == 3


def test_add_and_remove_class_attribute():
    from System import TimeSpan

    for _ in range(100):
        TimeSpan.new_method = lambda self_: self_.TotalMinutes
        ts = TimeSpan.FromHours(1)
        assert ts.new_method() == 60
        del TimeSpan.new_method
        assert not hasattr(ts, "new_method")


def test_comparisons():
    from System import DateTimeOffset
    from Python.Test import ClassTest

    d1 = DateTimeOffset.Parse("2016-11-14")
    d2 = DateTimeOffset.Parse("2016-11-15")

    assert (d1 == d2) == False
    assert (d1 != d2) == True

    assert (d1 < d2) == True
    assert (d1 <= d2) == True
    assert (d1 >= d2) == False
    assert (d1 > d2) == False

    assert (d1 == d1) == True
    assert (d1 != d1) == False

    assert (d1 < d1) == False
    assert (d1 <= d1) == True
    assert (d1 >= d1) == True
    assert (d1 > d1) == False

    assert (d2 == d1) == False
    assert (d2 != d1) == True

    assert (d2 < d1) == False
    assert (d2 <= d1) == False
    assert (d2 >= d1) == True
    assert (d2 > d1) == True

    with pytest.raises(TypeError):
        d1 < None

    with pytest.raises(TypeError):
        d1 < System.Guid()

    # ClassTest does not implement IComparable
    c1 = ClassTest()
    c2 = ClassTest()
    with pytest.raises(TypeError):
        c1 < c2


def test_self_callback():
    """Test calling back and forth between this and a c# baseclass."""

    class CallbackUser(Test.SelfCallbackTest):
        def DoCallback(self):
            self.PyCallbackWasCalled = False
            self.SameReference = False
            return self.Callback(self)

        def PyCallback(self, self2):
            self.PyCallbackWasCalled = True
            self.SameReference = self == self2

    testobj = CallbackUser()
    testobj.DoCallback()
    assert testobj.PyCallbackWasCalled
    assert testobj.SameReference


def test_method_inheritance():
    """Ensure that we call the overridden method instead of the one provided in
       the base class."""

    base = Test.BaseClass()
    derived = Test.DerivedClass()

    assert base.IsBase() == True
    assert derived.IsBase() == False


def test_callable():
    """Test that only delegate subtypes are callable"""

    def foo():
        pass

    assert callable(System.String("foo")) == False
    assert callable(System.Action(foo)) == True



_UNSET = object()

class PyErr:
    def __init__(self, type=_UNSET, value=_UNSET, traceback=_UNSET):
        if not(type is _UNSET):
            self.type = type
        if not(value is _UNSET):
            self.value = value
        if not(traceback is _UNSET):
            self.traceback = traceback



import clr
import sys
if sys.platform.lower() not in ['cli','win32']:
    print("only windows is supported for wpf")
clr.AddReference(r"wpf\PresentationFramework")
from System.IO import StreamReader
from System.Windows.Markup import XamlReader
from System.Threading import Thread, ThreadStart, ApartmentState
from System.Windows import Application, Window


class MyWindow(Window):
    def __init__(self):
        stream = StreamReader("DynamicGrid.xaml")
        window = XamlReader.Load(stream.BaseStream)
        Application().Run(window)
            

if __name__ == '__main__':
    thread = Thread(ThreadStart(MyWindow))
    thread.SetApartmentState(ApartmentState.STA)
    thread.Start()
    thread.Join()



"""
Code in this module gets loaded into the main clr module.
"""

__version__ = "3.0.0dev"


class clrproperty(object):
    """
    Property decorator for exposing python properties to .NET.
    The property type must be specified as the only argument to clrproperty.

    e.g.::

        class X(object):
            @clrproperty(string)
            def test(self):
                return "x"

    Properties decorated this way can be called from .NET, e.g.::

        dynamic x = getX(); // get an instance of X declared in Python
        string z = x.test; // calls into python and returns "x"
    """

    def __init__(self, type_, fget=None, fset=None):
        self.__name__ = getattr(fget, "__name__", None)
        self._clr_property_type_ = type_
        self.fget = fget
        self.fset = fset

    def __call__(self, fget):
        return self.__class__(self._clr_property_type_,
                              fget=fget,
                              fset=self.fset)

    def setter(self, fset):
        self.fset = fset
        return self

    def getter(self, fget):
        self.fget = fget
        return self

    def __get__(self, instance, owner):
        return self.fget.__get__(instance, owner)()

    def __set__(self, instance, value):
        if not self.fset:
            raise AttributeError("%s is read-only" % self.__name__)
        return self.fset.__get__(instance, None)(value)


class clrmethod(object):
    """
    Method decorator for exposing python methods to .NET.
    The argument and return types must be specified as arguments to clrmethod.

    e.g.::

        class X(object):
            @clrmethod(int, [str])
            def test(self, x):
                return len(x)

    Methods decorated this way can be called from .NET, e.g.::

        dynamic x = getX(); // get an instance of X declared in Python
        int z = x.test("hello"); // calls into python and returns len("hello")
    """

    def __init__(self, return_type, arg_types, clrname=None, func=None):
        self.__name__ = getattr(func, "__name__", None)
        self._clr_return_type_ = return_type
        self._clr_arg_types_ = arg_types
        self._clr_method_name_ = clrname or self.__name__
        self.__func = func

    def __call__(self, func):
        return self.__class__(self._clr_return_type_,
                              self._clr_arg_types_,
                              clrname=self._clr_method_name_,
                              func=func)

    def __get__(self, instance, owner):
        return self.__func.__get__(instance, owner)



# -*- coding: utf-8 -*-

"""Test CLR delegate support."""

import Python.Test as Test
import System
import pytest
from Python.Test import DelegateTest, StringDelegate

from .utils import HelloClass, hello_func, MultipleHandler, DictProxyType


def test_delegate_standard_attrs():
    """Test standard delegate attributes."""
    from Python.Test import PublicDelegate

    assert PublicDelegate.__name__ == 'PublicDelegate'
    assert PublicDelegate.__module__ == 'Python.Test'
    assert isinstance(PublicDelegate.__dict__, DictProxyType)
    assert PublicDelegate.__doc__ is None


def test_global_delegate_visibility():
    """Test visibility of module-level delegates."""
    from Python.Test import PublicDelegate

    assert PublicDelegate.__name__ == 'PublicDelegate'
    assert Test.PublicDelegate.__name__ == 'PublicDelegate'

    with pytest.raises(ImportError):
        from Python.Test import InternalDelegate
        _ = InternalDelegate

    with pytest.raises(AttributeError):
        _ = Test.InternalDelegate


def test_nested_delegate_visibility():
    """Test visibility of nested delegates."""
    ob = DelegateTest.PublicDelegate
    assert ob.__name__ == 'PublicDelegate'

    ob = DelegateTest.ProtectedDelegate
    assert ob.__name__ == 'ProtectedDelegate'

    with pytest.raises(AttributeError):
        _ = DelegateTest.InternalDelegate

    with pytest.raises(AttributeError):
        _ = DelegateTest.PrivateDelegate


def test_delegate_from_function():
    """Test delegate implemented with a Python function."""

    d = StringDelegate(hello_func)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_from_method():
    """Test delegate implemented with a Python instance method."""

    inst = HelloClass()
    d = StringDelegate(inst.hello)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_from_unbound_method():
    """Test failure mode for unbound methods."""

    with pytest.raises(TypeError):
        d = StringDelegate(HelloClass.hello)
        d()


def test_delegate_from_static_method():
    """Test delegate implemented with a Python static method."""

    d = StringDelegate(HelloClass.s_hello)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"

    inst = HelloClass()
    d = StringDelegate(inst.s_hello)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_from_class_method():
    """Test delegate implemented with a Python class method."""

    d = StringDelegate(HelloClass.c_hello)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"

    inst = HelloClass()
    d = StringDelegate(inst.c_hello)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_from_callable():
    """Test delegate implemented with a Python callable object."""

    inst = HelloClass()
    d = StringDelegate(inst)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_from_managed_instance_method():
    """Test delegate implemented with a managed instance method."""
    ob = DelegateTest()
    d = StringDelegate(ob.SayHello)

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_from_managed_static_method():
    """Test delegate implemented with a managed static method."""
    d = StringDelegate(DelegateTest.StaticSayHello)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d) == "hello"
    assert d() == "hello"

    ob.stringDelegate = d
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_from_delegate():
    """Test delegate implemented with another delegate."""
    d1 = StringDelegate(hello_func)
    d2 = StringDelegate(d1)
    ob = DelegateTest()

    assert ob.CallStringDelegate(d2) == "hello"
    assert d2() == "hello"

    ob.stringDelegate = d2
    assert ob.CallStringDelegate(ob.stringDelegate) == "hello"
    assert ob.stringDelegate() == "hello"


def test_delegate_with_invalid_args():
    """Test delegate instantiation with invalid (non-callable) args."""

    with pytest.raises(TypeError):
        _ = StringDelegate(None)

    with pytest.raises(TypeError):
        _ = StringDelegate("spam")

    with pytest.raises(TypeError):
        _ = StringDelegate(1)


def test_multicast_delegate():
    """Test multicast delegates."""

    inst = MultipleHandler()
    d1 = StringDelegate(inst.count)
    d2 = StringDelegate(inst.count)

    md = System.Delegate.Combine(d1, d2)
    ob = DelegateTest()

    assert ob.CallStringDelegate(md) == "ok"
    assert inst.value == 2

    assert md() == "ok"
    assert inst.value == 4


def test_subclass_delegate_fails():
    """Test that subclassing of a delegate type fails."""
    from Python.Test import PublicDelegate

    with pytest.raises(TypeError):
        class Boom(PublicDelegate):
            pass

        _ = Boom


def test_delegate_equality():
    """Test delegate equality."""

    d = StringDelegate(hello_func)
    ob = DelegateTest()
    ob.stringDelegate = d
    assert ob.stringDelegate == d


def test_bool_delegate():
    """Test boolean delegate."""
    from Python.Test import BoolDelegate

    def always_so_negative():
        return False

    d = BoolDelegate(always_so_negative)
    ob = DelegateTest()
    ob.CallBoolDelegate(d)

    assert not d()
    assert not ob.CallBoolDelegate(d)

    def always_so_positive():
        return 1
    bad = BoolDelegate(always_so_positive)
    with pytest.raises(TypeError):
        ob.CallBoolDelegate(bad)

def test_object_delegate():
    """Test object delegate."""
    from Python.Test import ObjectDelegate

    def create_object():
        return DelegateTest()

    d = ObjectDelegate(create_object)
    ob = DelegateTest()
    ob.CallObjectDelegate(d)

def test_invalid_object_delegate():
    """Test invalid object delegate with mismatched return type."""
    from Python.Test import ObjectDelegate

    d = ObjectDelegate(hello_func)
    ob = DelegateTest()
    with pytest.raises(TypeError):
        ob.CallObjectDelegate(d)

def test_out_int_delegate():
    """Test delegate with an out int parameter."""
    from Python.Test import OutIntDelegate
    value = 7

    def out_hello_func(ignored):
        return 5

    d = OutIntDelegate(out_hello_func)
    result = d(value)
    assert result == 5

    ob = DelegateTest()
    result = ob.CallOutIntDelegate(d, value)
    assert result == 5

    def invalid_handler(ignored):
        return '5'

    d = OutIntDelegate(invalid_handler)
    with pytest.raises(TypeError):
        result = d(value)

def test_out_string_delegate():
    """Test delegate with an out string parameter."""
    from Python.Test import OutStringDelegate
    value = 'goodbye'

    def out_hello_func(ignored):
        return 'hello'

    d = OutStringDelegate(out_hello_func)
    result = d(value)
    assert result == 'hello'

    ob = DelegateTest()
    result = ob.CallOutStringDelegate(d, value)
    assert result == 'hello'

def test_ref_int_delegate():
    """Test delegate with a ref string parameter."""
    from Python.Test import RefIntDelegate
    value = 7

    def ref_hello_func(data):
        assert data == value
        return data + 1

    d = RefIntDelegate(ref_hello_func)
    result = d(value)
    assert result == value + 1

    ob = DelegateTest()
    result = ob.CallRefIntDelegate(d, value)
    assert result == value + 1

def test_ref_string_delegate():
    """Test delegate with a ref string parameter."""
    from Python.Test import RefStringDelegate
    value = 'goodbye'

    def ref_hello_func(data):
        assert data == value
        return 'hello'

    d = RefStringDelegate(ref_hello_func)
    result = d(value)
    assert result == 'hello'

    ob = DelegateTest()
    result = ob.CallRefStringDelegate(d, value)
    assert result == 'hello'

def test_ref_int_ref_string_delegate():
    """Test delegate with a ref int and ref string parameter."""
    from Python.Test import RefIntRefStringDelegate
    intData = 7
    stringData = 'goodbye'

    def ref_hello_func(intValue, stringValue):
        assert intData == intValue
        assert stringData == stringValue
        return (intValue + 1, stringValue + '!')

    d = RefIntRefStringDelegate(ref_hello_func)
    result = d(intData, stringData)
    assert result == (intData + 1, stringData + '!')

    ob = DelegateTest()
    result = ob.CallRefIntRefStringDelegate(d, intData, stringData)
    assert result == (intData + 1, stringData + '!')

    def not_a_tuple(intValue, stringValue):
        return 'a'

    d = RefIntRefStringDelegate(not_a_tuple)
    with pytest.raises(TypeError):
        result = d(intData, stringData)

    def short_tuple(intValue, stringValue):
        return (5,)

    d = RefIntRefStringDelegate(short_tuple)
    with pytest.raises(TypeError):
        result = d(intData, stringData)

    def long_tuple(intValue, stringValue):
        return (5, 'a', 'b')

    d = RefIntRefStringDelegate(long_tuple)
    with pytest.raises(TypeError):
        result = d(intData, stringData)

    def wrong_tuple_item(intValue, stringValue):
        return ('a', 'b')

    d = RefIntRefStringDelegate(wrong_tuple_item)
    with pytest.raises(TypeError):
        result = d(intData, stringData)

def test_int_ref_int_ref_string_delegate():
    """Test delegate with a ref int and ref string parameter."""
    from Python.Test import IntRefIntRefStringDelegate
    intData = 7
    stringData = 'goodbye'

    def ref_hello_func(intValue, stringValue):
        assert intData == intValue
        assert stringData == stringValue
        return (intValue + len(stringValue), intValue + 1, stringValue + '!')

    d = IntRefIntRefStringDelegate(ref_hello_func)
    result = d(intData, stringData)
    assert result == (intData + len(stringData), intData + 1, stringData + '!')

    ob = DelegateTest()
    result = ob.CallIntRefIntRefStringDelegate(d, intData, stringData)
    assert result == (intData + len(stringData), intData + 1, stringData + '!')

    def not_a_tuple(intValue, stringValue):
        return 'a'

    d = IntRefIntRefStringDelegate(not_a_tuple)
    with pytest.raises(TypeError):
        result = d(intData, stringData)

    def short_tuple(intValue, stringValue):
        return (5,)

    d = IntRefIntRefStringDelegate(short_tuple)
    with pytest.raises(TypeError):
        result = d(intData, stringData)

    def wrong_return_type(intValue, stringValue):
        return ('a', 7, 'b')

    d = IntRefIntRefStringDelegate(wrong_return_type)
    with pytest.raises(TypeError):
        result = d(intData, stringData)

    # test async delegates

    # test multicast delegates

    # test explicit op_

    # test sig mismatch, both on managed and Python side

    # test return wrong type



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Run all of the unit tests for this package."""

from __future__ import print_function

import sys
import pytest

try:
    import System
except ImportError:
    print("Load clr import hook")
    import clr

    clr.AddReference("Python.Test")
    clr.AddReference("System.Collections")
    clr.AddReference("System.Data")
    clr.AddReference("System.Management")


def main(verbosity=1):
    # test_module passes on its own, but not here if
    # other test modules that import System.Windows.Forms
    # run first. They must not do module level import/AddReference()
    # of the System.Windows.Forms namespace.

    # FIXME: test_engine has tests that are being skipped.
    # FIXME: test_subclass has tests that are being skipped.
    pytest.main()


if __name__ == '__main__':
    main()
    if '--pause' in sys.argv:
        print("Press enter to continue")
        input()



# -*- coding: utf-8 -*-

"""Test CLR generics support."""

import clr

import System
import pytest


def assert_generic_wrapper_by_type(ptype, value):
    """Test Helper"""
    from Python.Test import GenericWrapper
    import System

    inst = GenericWrapper[ptype](value)
    assert inst.value == value

    atype = System.Array[ptype]
    items = atype([value, value, value])
    inst = GenericWrapper[atype](items)
    assert len(inst.value) == 3
    assert inst.value[0] == value
    assert inst.value[1] == value


def assert_generic_method_by_type(ptype, value, test_type=0):
    """Test Helper"""
    from Python.Test import GenericMethodTest, GenericStaticMethodTest
    import System

    itype = GenericMethodTest[System.Type]
    stype = GenericStaticMethodTest[System.Type]

    # Explicit selection (static method)
    result = stype.Overloaded[ptype](value)
    if test_type:
        assert result.__class__ == value.__class__
    else:
        assert result == value

    # Type inference (static method)
    result = stype.Overloaded(value)
    assert result == value
    if test_type:
        assert result.__class__ == value.__class__
    else:
        assert result == value

    # Explicit selection (instance method)
    result = itype().Overloaded[ptype](value)
    assert result == value
    if test_type:
        assert result.__class__ == value.__class__
    else:
        assert result == value

    # Type inference (instance method)
    result = itype().Overloaded(value)
    assert result == value
    if test_type:
        assert result.__class__ == value.__class__
    else:
        assert result == value

    atype = System.Array[ptype]
    items = atype([value, value, value])

    # Explicit selection (static method)
    result = stype.Overloaded[atype](items)
    if test_type:
        assert len(result) == 3
        assert result[0].__class__ == value.__class__
        assert result[1].__class__ == value.__class__
    else:
        assert len(result) == 3
        assert result[0] == value
        assert result[1] == value

    # Type inference (static method)
    result = stype.Overloaded(items)
    if test_type:
        assert len(result) == 3
        assert result[0].__class__ == value.__class__
        assert result[1].__class__ == value.__class__
    else:
        assert len(result) == 3
        assert result[0] == value
        assert result[1] == value

    # Explicit selection (instance method)
    result = itype().Overloaded[atype](items)
    if test_type:
        assert len(result) == 3
        assert result[0].__class__ == value.__class__
        assert result[1].__class__ == value.__class__
    else:
        assert len(result) == 3
        assert result[0] == value
        assert result[1] == value

    # Type inference (instance method)
    result = itype().Overloaded(items)
    if test_type:
        assert len(result) == 3
        assert result[0].__class__ == value.__class__
        assert result[1].__class__ == value.__class__
    else:
        assert len(result) == 3
        assert result[0] == value
        assert result[1] == value


def test_python_type_aliasing():
    """Test python type alias support with generics."""
    from System.Collections.Generic import Dictionary

    dict_ = Dictionary[str, str]()
    assert dict_.Count == 0
    dict_.Add("one", "one")
    assert dict_["one"] == "one"

    dict_ = Dictionary[System.String, System.String]()
    assert dict_.Count == 0
    dict_.Add("one", "one")
    assert dict_["one"] == "one"

    dict_ = Dictionary[int, int]()
    assert dict_.Count == 0
    dict_.Add(1, 1)
    assert dict_[1] == 1

    dict_ = Dictionary[System.Int32, System.Int32]()
    assert dict_.Count == 0
    dict_.Add(1, 1)
    assert dict_[1] == 1

    dict_ = Dictionary[int, int]()
    assert dict_.Count == 0
    dict_.Add(1, 1)
    assert dict_[1] == 1

    dict_ = Dictionary[System.Int64, System.Int64]()
    assert dict_.Count == 0
    dict_.Add(1, 1)
    assert dict_[1] == 1

    dict_ = Dictionary[float, float]()
    assert dict_.Count == 0
    dict_.Add(1.5, 1.5)
    assert dict_[1.5] == 1.5

    dict_ = Dictionary[System.Double, System.Double]()
    assert dict_.Count == 0
    dict_.Add(1.5, 1.5)
    assert dict_[1.5] == 1.5

    dict_ = Dictionary[bool, bool]()
    assert dict_.Count == 0
    dict_.Add(True, False)
    assert dict_[True] is False

    dict_ = Dictionary[System.Boolean, System.Boolean]()
    assert dict_.Count == 0
    dict_.Add(True, False)
    assert dict_[True] is False


def test_generic_reference_type():
    """Test usage of generic reference type definitions."""
    from Python.Test import GenericTypeDefinition

    inst = GenericTypeDefinition[System.String, System.Int32]("one", 2)
    assert inst.value1 == "one"
    assert inst.value2 == 2


def test_generic_value_type():
    """Test usage of generic value type definitions."""
    inst = System.Nullable[System.Int32](10)
    assert inst.HasValue
    assert inst.Value == 10


def test_generic_interface():
    # TODO NotImplemented
    pass


def test_generic_delegate():
    # TODO NotImplemented
    pass


def test_open_generic_type():
    """Test behavior of reflected open constructed generic types."""
    from Python.Test import DerivedFromOpenGeneric

    open_generic_type = DerivedFromOpenGeneric.__bases__[0]

    with pytest.raises(TypeError):
        _ = open_generic_type()

    with pytest.raises(TypeError):
        _ = open_generic_type[System.String]


def test_derived_from_open_generic_type():
    """Test a generic type derived from an open generic type."""
    from Python.Test import DerivedFromOpenGeneric

    type_ = DerivedFromOpenGeneric[System.String, System.String]
    inst = type_(1, 'two', 'three')

    assert inst.value1 == 1
    assert inst.value2 == 'two'
    assert inst.value3 == 'three'


def test_generic_type_name_resolution():
    """Test the ability to disambiguate generic type names."""
    from Python.Test import GenericNameTest1, GenericNameTest2

    # If both a non-generic and generic type exist for a name, the
    # unadorned name always resolves to the non-generic type.
    _class = GenericNameTest1
    assert _class().value == 0
    assert _class.value == 0

    # If no non-generic type exists for a name, the unadorned name
    # cannot be instantiated. It can only be used to bind a generic.

    with pytest.raises(TypeError):
        _ = GenericNameTest2()

    _class = GenericNameTest2[int]
    assert _class().value == 1
    assert _class.value == 1

    _class = GenericNameTest2[int, int]
    assert _class().value == 2
    assert _class.value == 2


def test_generic_type_binding():
    """Test argument conversion / binding for generic methods."""
    from Python.Test import InterfaceTest, ISayHello1, ShortEnum
    import System

    assert_generic_wrapper_by_type(System.Boolean, True)
    assert_generic_wrapper_by_type(bool, True)
    assert_generic_wrapper_by_type(System.Byte, 255)
    assert_generic_wrapper_by_type(System.SByte, 127)
    assert_generic_wrapper_by_type(System.Char, u'A')
    assert_generic_wrapper_by_type(System.Int16, 32767)
    assert_generic_wrapper_by_type(System.Int32, 2147483647)
    assert_generic_wrapper_by_type(int, 2147483647)
    assert_generic_wrapper_by_type(System.Int64, 9223372036854775807)
    assert_generic_wrapper_by_type(System.UInt16, 65000)
    assert_generic_wrapper_by_type(System.UInt32, 4294967295)
    assert_generic_wrapper_by_type(System.UInt64, 18446744073709551615)
    assert_generic_wrapper_by_type(System.Single, System.Single(3.402823e38))
    assert_generic_wrapper_by_type(System.Double, 1.7976931348623157e308)
    assert_generic_wrapper_by_type(float, 1.7976931348623157e308)
    assert_generic_wrapper_by_type(System.Decimal, System.Decimal.One)
    assert_generic_wrapper_by_type(System.String, "test")
    assert_generic_wrapper_by_type(str, "test")
    assert_generic_wrapper_by_type(ShortEnum, ShortEnum.Zero)
    assert_generic_wrapper_by_type(System.Object, InterfaceTest())
    assert_generic_wrapper_by_type(InterfaceTest, InterfaceTest())
    assert_generic_wrapper_by_type(ISayHello1, InterfaceTest())


def test_generic_method_binding():
    from Python.Test import GenericMethodTest, GenericStaticMethodTest
    from System import InvalidOperationException

    # Can invoke a static member on a closed generic type.
    value = GenericStaticMethodTest[str].Overloaded()
    assert value == 1

    with pytest.raises(InvalidOperationException):
        # Cannot invoke a static member on an open type.
        GenericStaticMethodTest.Overloaded()

    # Can invoke an instance member on a closed generic type.
    value = GenericMethodTest[str]().Overloaded()
    assert value == 1

    with pytest.raises(TypeError):
        # Cannot invoke an instance member on an open type,
        # because the open type cannot be instantiated.
        GenericMethodTest().Overloaded()


def test_generic_method_type_handling():
    """Test argument conversion / binding for generic methods."""
    from Python.Test import InterfaceTest, ISayHello1, ShortEnum
    import System

    # FIXME: Fails because Converter.GetTypeByAlias returns int32 for any integer, including ulong
    # assert_generic_method_by_type(System.UInt64, 18446744073709551615)
    assert_generic_method_by_type(System.Boolean, True)
    assert_generic_method_by_type(bool, True)
    assert_generic_method_by_type(System.Byte, 255)
    assert_generic_method_by_type(System.SByte, 127)
    assert_generic_method_by_type(System.Char, u'A')
    assert_generic_method_by_type(System.Int16, 32767)
    assert_generic_method_by_type(System.Int32, 2147483647)
    assert_generic_method_by_type(int, 2147483647)
    assert_generic_method_by_type(System.UInt16, 65000)
    assert_generic_method_by_type(System.Single, System.Single(3.402823e38))
    assert_generic_method_by_type(System.Double, 1.7976931348623157e308)
    assert_generic_method_by_type(float, 1.7976931348623157e308)
    assert_generic_method_by_type(System.Decimal, System.Decimal.One)
    assert_generic_method_by_type(System.String, "test")
    assert_generic_method_by_type(str, "test")
    assert_generic_method_by_type(ShortEnum, ShortEnum.Zero)
    assert_generic_method_by_type(System.Object, InterfaceTest())
    assert_generic_method_by_type(InterfaceTest, InterfaceTest(), 1)


def test_correct_overload_selection():
    """Test correct overloading selection for common types."""
    from System import (String, Double, Single,
                        Int16, Int32, Int64)
    from System import Math

    substr = String("substring")
    assert substr.Substring(2) == substr.Substring.__overloads__[Int32](
        Int32(2))
    assert substr.Substring(2, 3) == substr.Substring.__overloads__[Int32, Int32](
        Int32(2), Int32(3))

    for atype, value1, value2 in zip([Double, Single, Int16, Int32, Int64],
                                     [1.0, 1.0, 1, 1, 1],
                                     [2.0, 0.5, 2, 0, -1]):
        assert Math.Abs(atype(value1)) == Math.Abs.__overloads__[atype](atype(value1))
        assert Math.Abs(value1) == Math.Abs.__overloads__[atype](atype(value1))
        assert Math.Max(atype(value1),
                        atype(value2)) == Math.Max.__overloads__[atype, atype](
            atype(value1), atype(value2))
        assert Math.Max(atype(value1),
                        value2) == Math.Max.__overloads__[atype, atype](
            atype(value1), atype(value2))

    clr.AddReference("System.Runtime.InteropServices")
    from System.Runtime.InteropServices import GCHandle, GCHandleType
    from System import Array, Byte
    cs_array = Array.CreateInstance(Byte, 1000)
    handler = GCHandle.Alloc(cs_array, GCHandleType.Pinned)


def test_generic_method_overload_selection():
    """Test explicit overload selection with generic methods."""
    from Python.Test import GenericMethodTest, GenericStaticMethodTest

    type = GenericStaticMethodTest[str]
    inst = GenericMethodTest[str]()

    # public static int Overloaded()
    value = type.Overloaded()
    assert value == 1

    # public int Overloaded()
    value = inst.Overloaded()
    assert value == 1

    # public static T Overloaded(T arg) (inferred)
    value = type.Overloaded("test")
    assert value == "test"

    # public T Overloaded(T arg) (inferred)
    value = inst.Overloaded("test")
    assert value == "test"

    # public static T Overloaded(T arg) (explicit)
    value = type.Overloaded[str]("test")
    assert value == "test"

    # public T Overloaded(T arg) (explicit)
    value = inst.Overloaded[str]("test")
    assert value == "test"

    # public static Q Overloaded<Q>(Q arg)
    value = type.Overloaded[float](2.2)
    assert value == 2.2

    # public Q Overloaded<Q>(Q arg)
    value = inst.Overloaded[float](2.2)
    assert value == 2.2

    # public static Q Overloaded<Q>(Q arg)
    value = type.Overloaded[bool](True)
    assert value is True

    # public Q Overloaded<Q>(Q arg)
    value = inst.Overloaded[bool](True)
    assert value is True

    # public static U Overloaded<Q, U>(Q arg1, U arg2)
    value = type.Overloaded[bool, str](True, "true")
    assert value == "true"

    # public U Overloaded<Q, U>(Q arg1, U arg2)
    value = inst.Overloaded[bool, str](True, "true")
    assert value == "true"

    # public static U Overloaded<Q, U>(Q arg1, U arg2)
    value = type.Overloaded[str, bool]("true", True)
    assert value is True

    # public U Overloaded<Q, U>(Q arg1, U arg2)
    value = inst.Overloaded[str, bool]("true", True)
    assert value is True

    # public static string Overloaded<T>(int arg1, int arg2, string arg3)
    value = type.Overloaded[str](123, 456, "success")
    assert value == "success"

    # public string Overloaded<T>(int arg1, int arg2, string arg3)
    value = inst.Overloaded[str](123, 456, "success")
    assert value == "success"

    with pytest.raises(TypeError):
        _ = type.Overloaded[str, bool, int]("true", True, 1)

    with pytest.raises(TypeError):
        _ = inst.Overloaded[str, bool, int]("true", True, 1)


def test_method_overload_selection_with_generic_types():
    """Check method overload selection using generic types."""
    from Python.Test import ISayHello1, InterfaceTest, ShortEnum
    from Python.Test import MethodTest, GenericWrapper

    inst = InterfaceTest()

    vtype = GenericWrapper[System.Boolean]
    input_ = vtype(True)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value is True

    vtype = GenericWrapper[bool]
    input_ = vtype(True)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value is True

    vtype = GenericWrapper[System.Byte]
    input_ = vtype(255)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 255

    vtype = GenericWrapper[System.SByte]
    input_ = vtype(127)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 127

    vtype = GenericWrapper[System.Char]
    input_ = vtype(u'A')
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == u'A'

    vtype = GenericWrapper[System.Char]
    input_ = vtype(65535)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == chr(65535)

    vtype = GenericWrapper[System.Int16]
    input_ = vtype(32767)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 32767

    vtype = GenericWrapper[System.Int32]
    input_ = vtype(2147483647)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 2147483647

    vtype = GenericWrapper[int]
    input_ = vtype(2147483647)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 2147483647

    vtype = GenericWrapper[System.Int64]
    input_ = vtype(9223372036854775807)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 9223372036854775807

    vtype = GenericWrapper[System.UInt16]
    input_ = vtype(65000)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 65000

    vtype = GenericWrapper[System.UInt32]
    input_ = vtype(4294967295)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 4294967295

    vtype = GenericWrapper[System.UInt64]
    input_ = vtype(18446744073709551615)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 18446744073709551615

    vtype = GenericWrapper[System.Single]
    input_ = vtype(3.402823e38)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == System.Single(3.402823e38)

    vtype = GenericWrapper[System.Double]
    input_ = vtype(1.7976931348623157e308)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 1.7976931348623157e308

    vtype = GenericWrapper[float]
    input_ = vtype(1.7976931348623157e308)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == 1.7976931348623157e308

    vtype = GenericWrapper[System.Decimal]
    input_ = vtype(System.Decimal.One)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == System.Decimal.One

    vtype = GenericWrapper[System.String]
    input_ = vtype("spam")
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == "spam"

    vtype = GenericWrapper[str]
    input_ = vtype("spam")
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == "spam"

    vtype = GenericWrapper[ShortEnum]
    input_ = vtype(ShortEnum.Zero)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value == ShortEnum.Zero

    vtype = GenericWrapper[System.Object]
    input_ = vtype(inst)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value.__class__ == inst.__class__

    vtype = GenericWrapper[InterfaceTest]
    input_ = vtype(inst)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value.__class__ == inst.__class__

    iface_class = ISayHello1(inst).__class__
    vtype = GenericWrapper[ISayHello1]
    input_ = vtype(inst)
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value.value.__class__ == iface_class

    vtype = System.Array[GenericWrapper[int]]
    input_ = vtype([GenericWrapper[int](0), GenericWrapper[int](1)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 0
    assert value[1].value == 1


def test_overload_selection_with_arrays_of_generic_types():
    """Check overload selection using arrays of generic types."""
    from Python.Test import ISayHello1, InterfaceTest, ShortEnum
    from Python.Test import MethodTest, GenericWrapper

    inst = InterfaceTest()

    gtype = GenericWrapper[System.Boolean]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(True), gtype(True)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value is True
    assert value.Length == 2

    gtype = GenericWrapper[bool]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(True), gtype(True)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value is True
    assert value.Length == 2

    gtype = GenericWrapper[System.Byte]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(255), gtype(255)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 255
    assert value.Length == 2

    gtype = GenericWrapper[System.SByte]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(127), gtype(127)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 127
    assert value.Length == 2

    gtype = GenericWrapper[System.Char]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(u'A'), gtype(u'A')])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == u'A'
    assert value.Length == 2

    gtype = GenericWrapper[System.Char]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(65535), gtype(65535)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == chr(65535)
    assert value.Length == 2

    gtype = GenericWrapper[System.Int16]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(32767), gtype(32767)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 32767
    assert value.Length == 2

    gtype = GenericWrapper[System.Int32]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(2147483647), gtype(2147483647)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 2147483647
    assert value.Length == 2

    gtype = GenericWrapper[int]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(2147483647), gtype(2147483647)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 2147483647
    assert value.Length == 2

    gtype = GenericWrapper[System.Int64]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(9223372036854775807),
                    gtype(9223372036854775807)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 9223372036854775807
    assert value.Length == 2

    gtype = GenericWrapper[System.UInt16]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(65000), gtype(65000)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 65000
    assert value.Length == 2

    gtype = GenericWrapper[System.UInt32]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(4294967295), gtype(4294967295)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 4294967295
    assert value.Length == 2

    gtype = GenericWrapper[System.UInt64]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(18446744073709551615),
                    gtype(18446744073709551615)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 18446744073709551615
    assert value.Length == 2

    gtype = GenericWrapper[System.Single]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(3.402823e38), gtype(3.402823e38)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == System.Single(3.402823e38)
    assert value.Length == 2

    gtype = GenericWrapper[System.Double]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(1.7976931348623157e308),
                    gtype(1.7976931348623157e308)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 1.7976931348623157e308
    assert value.Length == 2

    gtype = GenericWrapper[float]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(1.7976931348623157e308),
                    gtype(1.7976931348623157e308)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == 1.7976931348623157e308
    assert value.Length == 2

    gtype = GenericWrapper[System.Decimal]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(System.Decimal.One),
                    gtype(System.Decimal.One)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == System.Decimal.One
    assert value.Length == 2

    gtype = GenericWrapper[System.String]
    vtype = System.Array[gtype]
    input_ = vtype([gtype("spam"), gtype("spam")])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == "spam"
    assert value.Length == 2

    gtype = GenericWrapper[str]
    vtype = System.Array[gtype]
    input_ = vtype([gtype("spam"), gtype("spam")])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == "spam"
    assert value.Length == 2

    gtype = GenericWrapper[ShortEnum]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(ShortEnum.Zero), gtype(ShortEnum.Zero)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value == ShortEnum.Zero
    assert value.Length == 2

    gtype = GenericWrapper[System.Object]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(inst), gtype(inst)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value.__class__ == inst.__class__
    assert value.Length == 2

    gtype = GenericWrapper[InterfaceTest]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(inst), gtype(inst)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value.__class__ == inst.__class__
    assert value.Length == 2

    iface_class = ISayHello1(inst).__class__
    gtype = GenericWrapper[ISayHello1]
    vtype = System.Array[gtype]
    input_ = vtype([gtype(inst), gtype(inst)])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].value.__class__ == iface_class
    assert value.Length == 2


def test_generic_overload_selection_magic_name_only():
    """Test using only __overloads__ to select on type & sig"""
    # TODO NotImplemented
    pass


def test_nested_generic_class():
    """Check nested generic classes."""
    # TODO NotImplemented
    pass

def test_missing_generic_type():
    from System.Collections import IList
    with pytest.raises(TypeError):
        IList[bool]

def test_invalid_generic_type_parameter():
    from Python.Test import GenericTypeWithConstraint
    with pytest.raises(TypeError):
        GenericTypeWithConstraint[System.Object]

def test_invalid_generic_method_type_parameter():
    from Python.Test import ConversionTest
    with pytest.raises(TypeError):
        ConversionTest.Echo[System.Object]

def test_generic_list_array_conversion():
    """Test conversion of lists to generic array arguments."""
    from Python.Test import GenericArrayConversionTest
    from Python.Test import Spam

    items = []
    for i in range(10):
        items.append(Spam(str(i)))

    result = GenericArrayConversionTest.EchoRange[Spam](items)
    assert result[0].__class__ == Spam
    assert len(result) == 10

    # Currently raises an exception because the correct generic type parameter is not inferred
    with pytest.raises(TypeError):
        result = GenericArrayConversionTest.EchoRange(items)
        assert result[0].__class__ == Spam
        assert len(result) == 10



# -*- coding: utf-8 -*-

"""Test __repr__ output"""

import System
import pytest
from Python.Test import ReprTest

def test_basic():
    """Test Point class which implements both ToString and __repr__ without inheritance"""
    ob = ReprTest.Point(1,2)
    # point implements ToString() and __repr__()
    assert ob.__repr__() == "Point(1,2)"
    assert str(ob) == "Python.Test.ReprTest+Point: X=1, Y=2"

def test_system_string():
    """Test system string"""
    ob = System.String("hello")
    assert str(ob) == "hello"
    assert "<System.String object at " in ob.__repr__()

def test_str_only():
    """Test class implementing ToString() but not __repr__()"""
    ob = ReprTest.Bar()
    assert str(ob) == "I implement ToString() but not __repr__()!"
    assert "<Python.Test.Bar object at " in ob.__repr__()

def test_hierarchy1():
    """Test inheritance hierarchy with base & middle class implementing ToString"""
    ob1 = ReprTest.BazBase()
    assert str(ob1) == "Base class implementing ToString()!"
    assert "<Python.Test.BazBase object at " in ob1.__repr__()

    ob2 = ReprTest.BazMiddle()
    assert str(ob2) == "Middle class implementing ToString()!"
    assert "<Python.Test.BazMiddle object at " in ob2.__repr__()

    ob3 = ReprTest.Baz()
    assert str(ob3) == "Middle class implementing ToString()!"
    assert "<Python.Test.Baz object at " in ob3.__repr__()

def bad_tostring():
    """Test ToString that can't be used by str()"""
    ob = ReprTest.Quux()
    assert str(ob) == "Python.Test.ReprTest+Quux"
    assert "<Python.Test.Quux object at " in ob.__repr__()

def bad_repr():
    """Test incorrect implementation of repr"""
    ob1 = ReprTest.QuuzBase()
    assert str(ob1) == "Python.Test.ReprTest+QuuzBase"
    assert "<Python.Test.QuuzBase object at " in ob.__repr__()
    
    ob2 = ReprTest.Quuz()
    assert str(ob2) == "Python.Test.ReprTest+Quuz"
    assert "<Python.Test.Quuz object at " in ob.__repr__()

    ob3 = ReprTest.Corge()
    with pytest.raises(Exception):
        str(ob3)
    with pytest.raises(Exception):
        ob3.__repr__()

    ob4 = ReprTest.Grault()
    with pytest.raises(Exception):
        str(ob4)
    with pytest.raises(Exception):
        ob4.__repr__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# FIXME: TypeError: 'EventBinding' object is not callable

from __future__ import print_function

import clr
import gc
import sys

import System

from .utils import (CallableHandler, ClassMethodHandler, GenericHandler,
                    HelloClass, StaticMethodHandler, VarCallableHandler,
                    VariableArgsHandler, hello_func)


class LeakTest(object):
    """A leak-check test for the objects implemented in the managed
       runtime. For each kind of object tested, memory should reach
       a particular level after warming up and stay essentially the
       same, net of minor fluctuation induced by GC."""

    def __init__(self):
        self.count = 50000
        self.quiet = 0
        self._ws = 0

    def notify(self, msg):
        if not self.quiet:
            print(msg)

    def start_test(self):
        System.GC.Collect(System.GC.MaxGeneration)
        gc.collect()
        self._ws = System.Environment.WorkingSet

    def end_test(self):
        start = self._ws
        System.GC.Collect(System.GC.MaxGeneration)
        gc.collect()
        end = System.Environment.WorkingSet
        diff = end - start
        if diff > 0:
            diff = '+{0}'.format(diff)
        else:
            diff = '{0}'.format(diff)
        print("  start: {0}  end: {1} diff: {2}".format(start, end, diff))
        print("")

    def run(self):
        self.testModules()
        self.testClasses()
        self.testEnumerations()
        self.testEvents()
        self.testDelegates()

    def report(self):
        gc.collect()
        dicttype = type({})
        for item in gc.get_objects():
            if type(item) != dicttype:
                print(item, sys.getrefcount(item))

    def test_modules(self):
        self.notify("Running module leak check...")

        for i in range(self.count):
            if i == 10:
                self.start_test()

            __import__('clr')
            __import__('System')
            __import__('System.IO')
            __import__('System.Net')
            __import__('System.Xml')

        self.end_test()

    def test_classes(self):
        from System.Collections import Hashtable
        from Python.Test import StringDelegate

        self.notify("Running class leak check...")

        for i in range(self.count):
            if i == 10:
                self.start_test()

            # Reference type
            x = Hashtable()
            del x

            # Value type
            x = System.Int32(99)
            del x

            # Delegate type
            x = StringDelegate(hello_func)
            del x

        self.end_test()

    def test_enumerations(self):
        import Python.Test as Test

        self.notify("Running enum leak check...")

        for i in range(self.count):
            if i == 10:
                self.start_test()

            x = Test.ByteEnum.Zero
            del x

            x = Test.SByteEnum.Zero
            del x

            x = Test.ShortEnum.Zero
            del x

            x = Test.UShortEnum.Zero
            del x

            x = Test.IntEnum.Zero
            del x

            x = Test.UIntEnum.Zero
            del x

            x = Test.LongEnum.Zero
            del x

            x = Test.ULongEnum.Zero
            del x

        self.end_test()

    def test_events(self):
        from Python.Test import EventTest, EventArgsTest

        self.notify("Running event leak check...")

        for i in range(self.count):
            if i == 10:
                self.start_test()

            testob = EventTest()

            # Instance method event handler
            handler = GenericHandler()
            testob.PublicEvent += handler.handler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= handler.handler
            del handler

            # Vararg method event handler
            handler = VariableArgsHandler()
            testob.PublicEvent += handler.handler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= handler.handler
            del handler

            # Callable object event handler
            handler = CallableHandler()
            testob.PublicEvent += handler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= handler
            del handler

            # Callable vararg event handler
            handler = VarCallableHandler()
            testob.PublicEvent += handler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= handler
            del handler

            # Static method event handler
            handler = StaticMethodHandler()
            StaticMethodHandler.value = None
            testob.PublicEvent += handler.handler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= handler.handler
            del handler

            # Class method event handler
            handler = ClassMethodHandler()
            ClassMethodHandler.value = None
            testob.PublicEvent += handler.handler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= handler.handler
            del handler

            # Managed instance event handler
            testob.PublicEvent += testob.GenericHandler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= testob.GenericHandler

            # Static managed event handler
            testob.PublicEvent += EventTest.StaticHandler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= EventTest.StaticHandler

            # Function event handler
            dict_ = {'value': None}

            def handler(sender, args, dict_=dict_):
                dict_['value'] = args.value

            testob.PublicEvent += handler
            testob.PublicEvent(testob, EventArgsTest(10))
            testob.PublicEvent -= handler
            del handler

        self.end_test()

    def test_delegates(self):
        from Python.Test import DelegateTest, StringDelegate

        self.notify("Running delegate leak check...")

        for i in range(self.count):
            if i == 10:
                self.start_test()

            # Delegate from function
            testob = DelegateTest()
            d = StringDelegate(hello_func)
            testob.CallStringDelegate(d)
            testob.stringDelegate = d
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del d

            # Delegate from instance method
            inst = HelloClass()
            testob = DelegateTest()
            d = StringDelegate(inst.hello)
            testob.CallStringDelegate(d)
            testob.stringDelegate = d
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del inst
            del d

            # Delegate from static method
            testob = DelegateTest()
            d = StringDelegate(HelloClass.s_hello)
            testob.CallStringDelegate(d)
            testob.stringDelegate = d
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del d

            # Delegate from class method
            testob = DelegateTest()
            d = StringDelegate(HelloClass.c_hello)
            testob.CallStringDelegate(d)
            testob.stringDelegate = d
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del d

            # Delegate from callable object
            inst = HelloClass()
            testob = DelegateTest()
            d = StringDelegate(inst)
            testob.CallStringDelegate(d)
            testob.stringDelegate = d
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del inst
            del d

            # Delegate from managed instance method
            testob = DelegateTest()
            d = StringDelegate(testob.SayHello)
            testob.CallStringDelegate(d)
            testob.stringDelegate = d
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del d

            # Delegate from managed static method
            testob = DelegateTest()
            d = StringDelegate(DelegateTest.StaticSayHello)
            testob.CallStringDelegate(d)
            testob.stringDelegate = d
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del d

            # Nested delegates
            testob = DelegateTest()
            d1 = StringDelegate(hello_func)
            d2 = StringDelegate(d1)
            testob.CallStringDelegate(d2)
            testob.stringDelegate = d2
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del d1
            del d2

            # Multicast delegates
            testob = DelegateTest()
            d1 = StringDelegate(hello_func)
            d2 = StringDelegate(hello_func)
            md = System.Delegate.Combine(d1, d2)
            testob.CallStringDelegate(md)
            testob.stringDelegate = md
            testob.stringDelegate()
            testob.stringDelegate = None
            del testob
            del d1
            del d2
            del md

        self.end_test()


if __name__ == '__main__':
    test = LeakTest()
    test.run()
    test.report()



# -*- coding: utf-8 -*-

"""Test CLR interface support."""

import Python.Test as Test
import pytest

from .utils import DictProxyType


def test_interface_standard_attrs():
    """Test standard class attributes."""
    from Python.Test import IPublicInterface

    assert IPublicInterface.__name__ == 'IPublicInterface'
    assert IPublicInterface.__module__ == 'Python.Test'
    assert isinstance(IPublicInterface.__dict__, DictProxyType)


def test_global_interface_visibility():
    """Test visibility of module-level interfaces."""
    from Python.Test import IPublicInterface

    assert IPublicInterface.__name__ == 'IPublicInterface'

    with pytest.raises(ImportError):
        from Python.Test import IInternalInterface
        _ = IInternalInterface

    with pytest.raises(AttributeError):
        _ = Test.IInternalInterface


def test_nested_interface_visibility():
    """Test visibility of nested interfaces."""
    from Python.Test import InterfaceTest

    ob = InterfaceTest.IPublic
    assert ob.__name__ == 'IPublic'

    ob = InterfaceTest.IProtected
    assert ob.__name__ == 'IProtected'

    with pytest.raises(AttributeError):
        _ = InterfaceTest.IInternal

    with pytest.raises(AttributeError):
        _ = InterfaceTest.IPrivate


def test_explicit_cast_to_interface():
    """Test explicit cast to an interface."""
    from Python.Test import InterfaceTest

    ob = InterfaceTest()
    assert type(ob).__name__ == 'InterfaceTest'
    assert hasattr(ob, 'HelloProperty')

    i1 = Test.ISayHello1(ob)
    assert type(i1).__name__ == 'ISayHello1'
    assert hasattr(i1, 'SayHello')
    assert i1.SayHello() == 'hello 1'
    assert not hasattr(i1, 'HelloProperty')
    assert i1.__implementation__ == ob
    assert i1.__raw_implementation__ == ob

    i2 = Test.ISayHello2(ob)
    assert type(i2).__name__ == 'ISayHello2'
    assert i2.SayHello() == 'hello 2'
    assert hasattr(i2, 'SayHello')
    assert not hasattr(i2, 'HelloProperty')


def test_interface_object_returned_through_method():
    """Test interface type is used if method return type is interface"""
    from Python.Test import InterfaceTest

    ob = InterfaceTest()
    hello1 = ob.GetISayHello1()
    assert type(hello1).__name__ == 'ISayHello1'
    assert hello1.__implementation__.__class__.__name__ == "InterfaceTest"

    assert hello1.SayHello() == 'hello 1'


def test_interface_object_returned_through_out_param():
    """Test interface type is used for out parameters of interface types"""
    from Python.Test import InterfaceTest

    ob = InterfaceTest()
    hello2 = ob.GetISayHello2(None)
    assert type(hello2).__name__ == 'ISayHello2'

    assert hello2.SayHello() == 'hello 2'


def test_null_interface_object_returned():
    """Test None is used also for methods with interface return types"""
    from Python.Test import InterfaceTest

    ob = InterfaceTest()
    hello1, hello2 = ob.GetNoSayHello(None)
    assert hello1 is None
    assert hello2 is None

def test_interface_array_returned():
    """Test interface type used for methods returning interface arrays"""
    from Python.Test import InterfaceTest

    ob = InterfaceTest()
    hellos = ob.GetISayHello1Array()
    assert type(hellos[0]).__name__ == 'ISayHello1'
    assert hellos[0].__implementation__.__class__.__name__ == "InterfaceTest"

def test_implementation_access():
    """Test the __implementation__ and __raw_implementation__ properties"""
    import System
    clrVal =  System.Int32(100)
    i = System.IComparable(clrVal)
    assert 100 == i.__implementation__
    assert clrVal == i.__raw_implementation__
    assert i.__implementation__ != i.__raw_implementation__


def test_interface_collection_iteration():
    """Test interface type is used when iterating over interface collection"""
    import System
    from System.Collections.Generic import List
    elem = System.IComparable(System.Int32(100))
    typed_list = List[System.IComparable]()
    typed_list.Add(elem)
    for e in typed_list:
        assert type(e).__name__ == "IComparable"

    untyped_list = System.Collections.ArrayList()
    untyped_list.Add(elem)
    for e in untyped_list:
        assert type(e).__name__ == "int"


def test_methods_of_Object_are_available():
    """Test calling methods inherited from Object"""
    import System
    clrVal =  System.Int32(100)
    i = System.IComparable(clrVal)
    assert i.Equals(clrVal)
    assert clrVal.GetHashCode() == i.GetHashCode()
    assert clrVal.GetType() == i.GetType()
    assert clrVal.ToString() == i.ToString()



# -*- coding: utf-8 -*-

"""Test support for managed arrays."""

import clr
import Python.Test as Test
import System
import pytest

from collections import UserList
from System import Single as float32


def test_public_array():
    """Test public arrays."""
    ob = Test.PublicArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    items[0] = 8
    assert items[0] == 8

    items[4] = 9
    assert items[4] == 9

    items[-4] = 0
    assert items[-4] == 0

    items[-1] = 4
    assert items[-1] == 4


def test_protected_array():
    """Test protected arrays."""
    ob = Test.ProtectedArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    items[0] = 8
    assert items[0] == 8

    items[4] = 9
    assert items[4] == 9

    items[-4] = 0
    assert items[-4] == 0

    items[-1] = 4
    assert items[-1] == 4


def test_internal_array():
    """Test internal arrays."""

    with pytest.raises(AttributeError):
        ob = Test.InternalArrayTest()
        _ = ob.items


def test_private_array():
    """Test private arrays."""

    with pytest.raises(AttributeError):
        ob = Test.PrivateArrayTest()
        _ = ob.items


def test_array_bounds_checking():
    """Test array bounds checking."""

    ob = Test.Int32ArrayTest()
    items = ob.items

    assert items[0] == 0
    assert items[1] == 1
    assert items[2] == 2
    assert items[3] == 3
    assert items[4] == 4

    assert items[-5] == 0
    assert items[-4] == 1
    assert items[-3] == 2
    assert items[-2] == 3
    assert items[-1] == 4

    with pytest.raises(IndexError):
        ob = Test.Int32ArrayTest()
        _ = ob.items[5]

    with pytest.raises(IndexError):
        ob = Test.Int32ArrayTest()
        ob.items[5] = 0

    with pytest.raises(IndexError):
        ob = Test.Int32ArrayTest()
        items[-6]

    with pytest.raises(IndexError):
        ob = Test.Int32ArrayTest()
        items[-6] = 0


def test_array_contains():
    """Test array support for __contains__."""

    ob = Test.Int32ArrayTest()
    items = ob.items

    assert 0 in items
    assert 1 in items
    assert 2 in items
    assert 3 in items
    assert 4 in items

    assert not (5 in items)  # "H:\Python27\Lib\unittest\case.py", line 592, in deprecated_func,
    assert not (-1 in items)  # TypeError: int() argument must be a string or a number, not 'NoneType'
    assert not (None in items)  # which threw ^ here which is a little odd.
    # But when run from runtests.py. Not when this module ran by itself.


def test_boolean_array():
    """Test boolean arrays."""
    ob = Test.BooleanArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] is True
    assert items[1] is False
    assert items[2] is True
    assert items[3] is False
    assert items[4] is True

    items[0] = False
    assert items[0] is False

    items[0] = True
    assert items[0] is True

    with pytest.raises(TypeError):
        ob = Test.ByteArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.ByteArrayTest()
        ob[0] = "wrong"


def test_byte_array():
    """Test byte arrays."""
    ob = Test.ByteArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 255
    min_ = 0

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.ByteArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.ByteArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.ByteArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.ByteArrayTest()
        ob[0] = "wrong"


def test_sbyte_array():
    """Test sbyte arrays."""
    ob = Test.SByteArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 127
    min_ = -128

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.SByteArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.SByteArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.SByteArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.SByteArrayTest()
        ob[0] = "wrong"


def test_char_array():
    """Test char arrays."""
    ob = Test.CharArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 'a'
    assert items[4] == 'e'

    max_ = chr(65535)
    min_ = chr(0)

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(TypeError):
        ob = Test.CharArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.CharArrayTest()
        ob[0] = "wrong"


def test_int16_array():
    """Test Int16 arrays."""
    ob = Test.Int16ArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 32767
    min_ = -32768

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.Int16ArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.Int16ArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.Int16ArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.Int16ArrayTest()
        ob[0] = "wrong"


def test_int32_array():
    """Test Int32 arrays."""
    ob = Test.Int32ArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 2147483647
    min_ = -2147483648

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.Int32ArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.Int32ArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.Int32ArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.Int32ArrayTest()
        ob[0] = "wrong"


def test_int64_array():
    """Test Int64 arrays."""
    ob = Test.Int64ArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 9223372036854775807
    min_ = -9223372036854775808

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.Int64ArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.Int64ArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.Int64ArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.Int64ArrayTest()
        ob[0] = "wrong"


def test_uint16_array():
    """Test UInt16 arrays."""
    ob = Test.UInt16ArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 65535
    min_ = 0

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.UInt16ArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.UInt16ArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.UInt16ArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.UInt16ArrayTest()
        ob[0] = "wrong"


def test_uint32_array():
    """Test UInt32 arrays."""
    ob = Test.UInt32ArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 4294967295
    min_ = 0

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.UInt32ArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.UInt32ArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.UInt32ArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.UInt32ArrayTest()
        ob[0] = "wrong"


def test_uint64_array():
    """Test UInt64 arrays."""
    ob = Test.UInt64ArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0
    assert items[4] == 4

    max_ = 18446744073709551615
    min_ = 0

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(OverflowError):
        ob = Test.UInt64ArrayTest()
        ob.items[0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.UInt64ArrayTest()
        ob.items[0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.UInt64ArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.UInt64ArrayTest()
        ob[0] = "wrong"


def test_single_array():
    """Test Single arrays."""
    ob = Test.SingleArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0.0
    assert items[4] == 4.0

    max_ = float32(3.402823e38)
    min_ = float32(-3.402823e38)

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(TypeError):
        ob = Test.SingleArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.SingleArrayTest()
        ob[0] = "wrong"


def test_double_array():
    """Test Double arrays."""
    ob = Test.DoubleArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == 0.0
    assert items[4] == 4.0

    max_ = 1.7976931348623157e308
    min_ = -1.7976931348623157e308

    items[0] = max_
    assert items[0] == max_

    items[0] = min_
    assert items[0] == min_

    items[-4] = max_
    assert items[-4] == max_

    items[-1] = min_
    assert items[-1] == min_

    with pytest.raises(TypeError):
        ob = Test.DoubleArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.DoubleArrayTest()
        ob[0] = "wrong"


def test_decimal_array():
    """Test Decimal arrays."""
    ob = Test.DecimalArrayTest()
    items = ob.items

    from System import Decimal
    max_d = Decimal.Parse("79228162514264337593543950335")
    min_d = Decimal.Parse("-79228162514264337593543950335")

    assert len(items) == 5

    assert items[0] == Decimal(0)
    assert items[4] == Decimal(4)

    items[0] = max_d
    assert items[0] == max_d

    items[0] = min_d
    assert items[0] == min_d

    items[-4] = max_d
    assert items[-4] == max_d

    items[-1] = min_d
    assert items[-1] == min_d

    with pytest.raises(TypeError):
        ob = Test.DecimalArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.DecimalArrayTest()
        ob[0] = "wrong"


def test_string_array():
    """Test String arrays."""
    ob = Test.StringArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == '0'
    assert items[4] == '4'

    items[0] = "spam"
    assert items[0] == "spam"

    items[0] = "eggs"
    assert items[0] == "eggs"

    items[-4] = "spam"
    assert items[-4] == "spam"

    items[-1] = "eggs"
    assert items[-1] == "eggs"

    with pytest.raises(TypeError):
        ob = Test.StringArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.Int64ArrayTest()
        ob[0] = 0


def test_enum_array():
    """Test enum arrays."""
    from Python.Test import ShortEnum
    ob = Test.EnumArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] == ShortEnum.Zero
    assert items[4] == ShortEnum.Four

    items[0] = ShortEnum.Four
    assert items[0] == ShortEnum.Four

    items[0] = ShortEnum.Zero
    assert items[0] == ShortEnum.Zero

    items[-4] = ShortEnum.Four
    assert items[-4] == ShortEnum.Four

    items[-1] = ShortEnum.Zero
    assert items[-1] == ShortEnum.Zero

    with pytest.raises(TypeError):
        ob = Test.EnumArrayTest()
        ob.items[0] = 99

    with pytest.raises(TypeError):
        ob = Test.EnumArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.EnumArrayTest()
        ob[0] = "wrong"


def test_object_array():
    """Test ob arrays."""
    from Python.Test import Spam
    ob = Test.ObjectArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0].GetValue() == "0"
    assert items[4].GetValue() == "4"

    items[0] = Spam("4")
    assert items[0].GetValue() == "4"

    items[0] = Spam("0")
    assert items[0].GetValue() == "0"

    items[-4] = Spam("4")
    assert items[-4].GetValue() == "4"

    items[-1] = Spam("0")
    assert items[-1].GetValue() == "0"

    items[0] = 99
    assert items[0] == 99

    items[0] = None
    assert items[0] is None

    with pytest.raises(TypeError):
        ob = Test.ObjectArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.ObjectArrayTest()
        ob.items["wrong"] = "wrong"


def test_null_array():
    """Test null arrays."""
    ob = Test.NullArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0] is None
    assert items[4] is None

    items[0] = "spam"
    assert items[0] == "spam"

    items[0] = None
    assert items[0] is None

    items[-4] = "spam"
    assert items[-4] == "spam"

    items[-1] = None
    assert items[-1] is None

    empty = ob.empty
    assert len(empty) == 0

    with pytest.raises(TypeError):
        ob = Test.NullArrayTest()
        _ = ob.items["wrong"]


def test_interface_array():
    """Test interface arrays."""
    from Python.Test import Spam
    ob = Test.InterfaceArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0].GetValue() == "0"
    assert items[4].GetValue() == "4"

    items[0] = Spam("4")
    assert items[0].GetValue() == "4"

    items[0] = Spam("0")
    assert items[0].GetValue() == "0"

    items[-4] = Spam("4")
    assert items[-4].GetValue() == "4"

    items[-1] = Spam("0")
    assert items[-1].GetValue() == "0"

    items[0] = None
    assert items[0] is None

    with pytest.raises(TypeError):
        ob = Test.InterfaceArrayTest()
        ob.items[0] = 99

    with pytest.raises(TypeError):
        ob = Test.InterfaceArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.InterfaceArrayTest()
        ob.items["wrong"] = "wrong"


def test_typed_array():
    """Test typed arrays."""
    from Python.Test import Spam
    ob = Test.TypedArrayTest()
    items = ob.items

    assert len(items) == 5

    assert items[0].GetValue() == "0"
    assert items[4].GetValue() == "4"

    items[0] = Spam("4")
    assert items[0].GetValue() == "4"

    items[0] = Spam("0")
    assert items[0].GetValue() == "0"

    items[-4] = Spam("4")
    assert items[-4].GetValue() == "4"

    items[-1] = Spam("0")
    assert items[-1].GetValue() == "0"

    items[0] = None
    assert items[0] is None

    with pytest.raises(TypeError):
        ob = Test.TypedArrayTest()
        ob.items[0] = 99

    with pytest.raises(TypeError):
        ob = Test.TypedArrayTest()
        _ = ob.items["wrong"]

    with pytest.raises(TypeError):
        ob = Test.TypedArrayTest()
        ob.items["wrong"] = Spam("0")

    with pytest.raises(TypeError):
        ob = Test.TypedArrayTest()
        _ = ob.items[0.5]

    with pytest.raises(TypeError):
        ob = Test.TypedArrayTest()
        ob.items[0.5] = Spam("0")

def test_multi_dimensional_array():
    """Test multi-dimensional arrays."""
    ob = Test.MultiDimensionalArrayTest()
    items = ob.items

    assert len(items) == 25

    assert items[0, 0] == 0
    assert items[0, 1] == 1
    assert items[0, 2] == 2
    assert items[0, 3] == 3
    assert items[0, 4] == 4
    assert items[1, 0] == 5
    assert items[1, 1] == 6
    assert items[1, 2] == 7
    assert items[1, 3] == 8
    assert items[1, 4] == 9
    assert items[2, 0] == 10
    assert items[2, 1] == 11
    assert items[2, 2] == 12
    assert items[2, 3] == 13
    assert items[2, 4] == 14
    assert items[3, 0] == 15
    assert items[3, 1] == 16
    assert items[3, 2] == 17
    assert items[3, 3] == 18
    assert items[3, 4] == 19
    assert items[4, 0] == 20
    assert items[4, 1] == 21
    assert items[4, 2] == 22
    assert items[4, 3] == 23
    assert items[4, 4] == 24

    max_ = 2147483647
    min_ = -2147483648

    items[0, 0] = max_
    assert items[0, 0] == max_

    items[0, 0] = min_
    assert items[0, 0] == min_

    items[-4, 0] = max_
    assert items[-4, 0] == max_

    items[-1, -1] = min_
    assert items[-1, -1] == min_

    with pytest.raises(OverflowError):
        ob = Test.MultiDimensionalArrayTest()
        ob.items[0, 0] = max_ + 1

    with pytest.raises(OverflowError):
        ob = Test.MultiDimensionalArrayTest()
        ob.items[0, 0] = min_ - 1

    with pytest.raises(TypeError):
        ob = Test.MultiDimensionalArrayTest()
        _ = ob.items["wrong", 0]

    with pytest.raises(TypeError):
        ob = Test.MultiDimensionalArrayTest()
        ob.items[0, 0] = "wrong"

    with pytest.raises(TypeError):
        ob = Test.MultiDimensionalArrayTest()
        ob["0", 0] = 0

    with pytest.raises(TypeError):
        ob = Test.MultiDimensionalArrayTest()
        ob.items["0", 0] = 0

    with pytest.raises(TypeError):
        ob = Test.MultiDimensionalArrayTest()
        _ = ob.items[0.5, 0]

    with pytest.raises(TypeError):
        ob = Test.MultiDimensionalArrayTest()
        ob.items[0.5, 0] = 0

def test_array_iteration():
    """Test array iteration."""
    items = Test.Int32ArrayTest().items

    for i in items:
        assert (i > -1) and (i < 5)

    items = Test.NullArrayTest().items

    for i in items:
        assert i is None

    empty = Test.NullArrayTest().empty

    for i in empty:
        raise TypeError('iteration over empty array')


def test_tuple_array_conversion():
    """Test conversion of tuples to array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    items = []
    for i in range(10):
        items.append(Spam(str(i)))
    items = tuple(items)

    result = ArrayConversionTest.EchoRange(items)
    assert result[0].__class__ == Spam
    assert len(result) == 10


def test_tuple_nested_array_conversion():
    """Test conversion of tuples to array-of-array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    items = []
    for i in range(10):
        subs = []
        for _ in range(10):
            subs.append(Spam(str(i)))
        items.append(tuple(subs))
    items = tuple(items)

    result = ArrayConversionTest.EchoRangeAA(items)

    assert len(result) == 10
    assert len(result[0]) == 10
    assert result[0][0].__class__ == Spam


def test_list_array_conversion():
    """Test conversion of lists to array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    items = []
    for i in range(10):
        items.append(Spam(str(i)))

    result = ArrayConversionTest.EchoRange(items)
    assert result[0].__class__ == Spam
    assert len(result) == 10


def test_list_nested_array_conversion():
    """Test conversion of lists to array-of-array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    items = []
    for i in range(10):
        subs = []
        for _ in range(10):
            subs.append(Spam(str(i)))
        items.append(subs)

    result = ArrayConversionTest.EchoRangeAA(items)

    assert len(result) == 10
    assert len(result[0]) == 10
    assert result[0][0].__class__ == Spam


def test_sequence_array_conversion():
    """Test conversion of sequence-like obs to array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    items = UserList()
    for i in range(10):
        items.append(Spam(str(i)))

    result = ArrayConversionTest.EchoRange(items)
    assert result[0].__class__ == Spam
    assert len(result) == 10


def test_sequence_nested_array_conversion():
    """Test conversion of sequences to array-of-array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    items = UserList()
    for i in range(10):
        subs = UserList()
        for _ in range(10):
            subs.append(Spam(str(i)))
        items.append(subs)

    result = ArrayConversionTest.EchoRangeAA(items)

    assert len(result) == 10
    assert len(result[0]) == 10
    assert result[0][0].__class__ == Spam


def test_tuple_array_conversion_type_checking():
    """Test error handling for tuple conversion to array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    # This should work, because null / None is a valid value in an
    # array of reference types.

    items = []
    for i in range(10):
        items.append(Spam(str(i)))
    items[1] = None
    items = tuple(items)

    result = ArrayConversionTest.EchoRange(items)

    assert result[0].__class__ == Spam
    assert result[1] is None
    assert len(result) == 10

    with pytest.raises(TypeError):
        temp = list(items)
        temp[1] = 1
        _ = ArrayConversionTest.EchoRange(tuple(temp))

    with pytest.raises(TypeError):
        temp = list(items)
        temp[1] = "spam"
        _ = ArrayConversionTest.EchoRange(tuple(temp))


def test_list_array_conversion_type_checking():
    """Test error handling for list conversion to array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    # This should work, because null / None is a valid value in an
    # array of reference types.

    items = []
    for i in range(10):
        items.append(Spam(str(i)))
    items[1] = None

    result = ArrayConversionTest.EchoRange(items)

    assert result[0].__class__ == Spam
    assert result[1] is None
    assert len(result) == 10

    with pytest.raises(TypeError):
        items[1] = 1
        _ = ArrayConversionTest.EchoRange(items)

    with pytest.raises(TypeError):
        items[1] = "spam"
        _ = ArrayConversionTest.EchoRange(items)


def test_sequence_array_conversion_type_checking():
    """Test error handling for sequence conversion to array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam

    # This should work, because null / None is a valid value in an
    # array of reference types.

    items = UserList()
    for i in range(10):
        items.append(Spam(str(i)))
    items[1] = None

    result = ArrayConversionTest.EchoRange(items)

    assert result[0].__class__ == Spam
    assert result[1] is None
    assert len(result) == 10

    with pytest.raises(TypeError):
        items[1] = 1
        _ = ArrayConversionTest.EchoRange(items)

    with pytest.raises(TypeError):
        items[1] = "spam"
        _ = ArrayConversionTest.EchoRange(items)


def test_md_array_conversion():
    """Test passing of multi-dimensional array arguments."""
    from Python.Test import ArrayConversionTest
    from Python.Test import Spam
    from System import Array

    # Currently, the runtime does not support automagic conversion of
    # Python sequences to true multi-dimensional arrays (though it
    # does support arrays-of-arrays). This test exists mostly as an
    # example of how a multi-dimensional array can be created and used
    # with managed code from Python.

    items = Array.CreateInstance(Spam, 5, 5)

    for i in range(5):
        for n in range(5):
            items.SetValue(Spam(str((i, n))), (i, n))

    result = ArrayConversionTest.EchoRangeMD(items)

    assert len(result) == 25
    assert result[0, 0].__class__ == Spam
    assert result[0, 0].__class__ == Spam


def test_boxed_value_type_mutation_result():
    """Test behavior of boxed value types."""

    # This test actually exists mostly as documentation of an important
    # concern when dealing with value types. Python does not have any
    # value type semantics that can be mapped to the CLR, so it is easy
    # to accidentally write code like the following which is not really
    # mutating value types in-place but changing boxed copies.

    from System import Array
    from Python.Test import Point

    items = Array.CreateInstance(Point, 5)

    for i in range(5):
        items[i] = Point(i, i)

    for i in range(5):
        # Boxed items, so set_attr will not change the array member.
        assert items[i].X == i
        assert items[i].Y == i
        items[i].X = i + 1
        items[i].Y = i + 1
        assert items[i].X == i
        assert items[i].Y == i

    for i in range(5):
        # Demonstrates the workaround that will change the members.
        assert items[i].X == i
        assert items[i].Y == i
        item = items[i]
        item.X = i + 1
        item.Y = i + 1
        items[i] = item
        assert items[i].X == i + 1
        assert items[i].Y == i + 1

def test_create_array_from_shape():
    from System import Array

    value = Array[int](3)
    assert value[1] == 0
    assert value.Length == 3

    value = Array[int](3, 4)
    assert value[1, 1] == 0
    assert value.GetLength(0) == 3
    assert value.GetLength(1) == 4

    with pytest.raises(ValueError):
        Array[int](-1)

    with pytest.raises(TypeError):
        Array[int]('1')

    with pytest.raises(ValueError):
        Array[int](-1, -1)

    with pytest.raises(TypeError):
        Array[int]('1', '1')

def test_special_array_creation():
    """Test using the Array[<type>] syntax for creating arrays."""
    from Python.Test import ISayHello1, InterfaceTest, ShortEnum
    from System import Array
    inst = InterfaceTest()

    value = Array[System.Boolean]([True, True])
    assert value[0] is True
    assert value[1] is True
    assert value.Length == 2

    value = Array[bool]([True, True])
    assert value[0] is True
    assert value[1] is True
    assert value.Length == 2

    value = Array[System.Byte]([0, 255])
    assert value[0] == 0
    assert value[1] == 255
    assert value.Length == 2

    value = Array[System.SByte]([0, 127])
    assert value[0] == 0
    assert value[1] == 127
    assert value.Length == 2

    value = Array[System.Char]([u'A', u'Z'])
    assert value[0] == u'A'
    assert value[1] == u'Z'
    assert value.Length == 2

    value = Array[System.Char]([0, 65535])
    assert value[0] == chr(0)
    assert value[1] == chr(65535)
    assert value.Length == 2

    value = Array[System.Int16]([0, 32767])
    assert value[0] == 0
    assert value[1] == 32767
    assert value.Length == 2

    value = Array[System.Int32]([0, 2147483647])
    assert value[0] == 0
    assert value[1] == 2147483647
    assert value.Length == 2

    value = Array[int]([0, 2147483647])
    assert value[0] == 0
    assert value[1] == 2147483647
    assert value.Length == 2

    value = Array[System.Int64]([0, 9223372036854775807])
    assert value[0] == 0
    assert value[1] == 9223372036854775807
    assert value.Length == 2

    value = Array[System.UInt16]([0, 65000])
    assert value[0] == 0
    assert value[1] == 65000
    assert value.Length == 2

    value = Array[System.UInt32]([0, 4294967295])
    assert value[0] == 0
    assert value[1] == 4294967295
    assert value.Length == 2

    value = Array[System.UInt64]([0, 18446744073709551615])
    assert value[0] == 0
    assert value[1] == 18446744073709551615
    assert value.Length == 2

    value = Array[System.Single]([0.0, 3.402823e38])
    assert value[0] == 0.0
    assert value[1] == System.Single(3.402823e38)
    assert value.Length == 2

    value = Array[System.Double]([0.0, 1.7976931348623157e308])
    assert value[0] == 0.0
    assert value[1] == 1.7976931348623157e308
    assert value.Length == 2

    value = Array[float]([0.0, 1.7976931348623157e308])
    assert value[0] == 0.0
    assert value[1] == 1.7976931348623157e308
    assert value.Length == 2

    value = Array[System.Decimal]([System.Decimal.Zero, System.Decimal.One])
    assert value[0] == System.Decimal.Zero
    assert value[1] == System.Decimal.One
    assert value.Length == 2

    value = Array[System.String](["one", "two"])
    assert value[0] == "one"
    assert value[1] == "two"
    assert value.Length == 2

    value = Array[str](["one", "two"])
    assert value[0] == "one"
    assert value[1] == "two"
    assert value.Length == 2

    value = Array[ShortEnum]([ShortEnum.Zero, ShortEnum.One])
    assert value[0] == ShortEnum.Zero
    assert value[1] == ShortEnum.One
    assert value.Length == 2

    value = Array[System.Object]([inst, inst])
    assert value[0].__class__ == inst.__class__
    assert value[1].__class__ == inst.__class__
    assert value.Length == 2

    value = Array[InterfaceTest]([inst, inst])
    assert value[0].__class__ == inst.__class__
    assert value[1].__class__ == inst.__class__
    assert value.Length == 2

    iface_class = ISayHello1(inst).__class__
    value = Array[ISayHello1]([inst, inst])
    assert value[0].__class__ == iface_class
    assert value[1].__class__ == iface_class
    assert value.Length == 2

    inst = System.Exception("badness")
    value = Array[System.Exception]([inst, inst])
    assert value[0].__class__ == inst.__class__
    assert value[1].__class__ == inst.__class__
    assert value.Length == 2


def test_array_abuse():
    """Test array abuse."""
    _class = Test.PublicArrayTest
    ob = Test.PublicArrayTest()

    with pytest.raises(AttributeError):
        del _class.__getitem__

    with pytest.raises(AttributeError):
        del ob.__getitem__

    with pytest.raises(AttributeError):
        del _class.__setitem__

    with pytest.raises(AttributeError):
        del ob.__setitem__

    with pytest.raises(TypeError):
        Test.PublicArrayTest.__getitem__(0, 0)

    with pytest.raises(AttributeError):
        Test.PublicArrayTest.__setitem__(0, 0, 0)

    with pytest.raises(KeyError):
        Test.PublicArrayTest.__dict__['__getitem__']

    with pytest.raises(KeyError):
        Test.PublicArrayTest.__dict__['__setitem__']


def test_iterator_to_array():
    from System import Array, String

    d = {"a": 1, "b": 2, "c": 3}
    keys_iterator = iter(d.keys())
    arr = Array[String](keys_iterator)

    Array.Sort(arr)

    assert arr[0] == "a"
    assert arr[1] == "b"
    assert arr[2] == "c"


def test_dict_keys_to_array():
    from System import Array, String

    d = {"a": 1, "b": 2, "c": 3}
    d_keys = d.keys()
    arr = Array[String](d_keys)

    Array.Sort(arr)

    assert arr[0] == "a"
    assert arr[1] == "b"
    assert arr[2] == "c"




# -*- coding: utf-8 -*-

"""Test clrmethod and clrproperty support for calling methods and getting/setting python properties from CLR."""

import Python.Test as Test
import System
import pytest
import clr

class ExampleClrClass(System.Object):
    __namespace__ = "PyTest"
    def __init__(self):
        self._x = 3
    @clr.clrmethod(int, [int])
    def test(self, x):
        return x*2

    def get_X(self):
        return self._x
    def set_X(self, value):
        self._x = value
    X = clr.clrproperty(int, get_X, set_X)

    @clr.clrproperty(int)
    def Y(self):
        return self._x * 2

def test_set_and_get_property_from_py():
    """Test setting and getting clr-accessible properties from python."""
    t = ExampleClrClass()
    assert t.X == 3
    assert t.Y == 3 * 2
    t.X = 4
    assert t.X == 4
    assert t.Y == 4 * 2

def test_set_and_get_property_from_clr():
    """Test setting and getting clr-accessible properties from the clr."""
    t = ExampleClrClass()
    assert t.GetType().GetProperty("X").GetValue(t) == 3
    assert t.GetType().GetProperty("Y").GetValue(t) == 3 * 2
    t.GetType().GetProperty("X").SetValue(t, 4)
    assert t.GetType().GetProperty("X").GetValue(t) == 4
    assert t.GetType().GetProperty("Y").GetValue(t) == 4 * 2


def test_set_and_get_property_from_clr_and_py():
    """Test setting and getting clr-accessible properties alternatingly from the clr and from python."""
    t = ExampleClrClass()
    assert t.GetType().GetProperty("X").GetValue(t) == 3
    assert t.GetType().GetProperty("Y").GetValue(t) == 3 * 2
    assert t.X == 3
    assert t.Y == 3 * 2
    t.GetType().GetProperty("X").SetValue(t, 4)
    assert t.GetType().GetProperty("X").GetValue(t) == 4
    assert t.GetType().GetProperty("Y").GetValue(t) == 4 * 2
    assert t.X == 4
    assert t.Y == 4 * 2
    t.X = 5
    assert t.GetType().GetProperty("X").GetValue(t) == 5
    assert t.GetType().GetProperty("Y").GetValue(t) == 5 * 2
    assert t.X == 5
    assert t.Y == 5 * 2

def test_method_invocation_from_py():
    """Test calling a clr-accessible method from python."""
    t = ExampleClrClass()
    assert t.test(41) == 41*2

def test_method_invocation_from_clr():
    """Test calling a clr-accessible method from the clr."""
    t = ExampleClrClass()
    assert t.GetType().GetMethod("test").Invoke(t, [37]) == 37*2



# -*- coding: utf-8 -*-

"""Test doc strings support."""


def test_doc_with_ctor():
    from Python.Test import DocWithCtorTest

    assert DocWithCtorTest.__doc__ == 'DocWithCtorTest Class'
    assert DocWithCtorTest.TestMethod.__doc__ == 'DocWithCtorTest TestMethod'
    assert DocWithCtorTest.StaticTestMethod.__doc__ == 'DocWithCtorTest StaticTestMethod'


def test_doc_with_ctor_no_doc():
    from Python.Test import DocWithCtorNoDocTest

    assert DocWithCtorNoDocTest.__doc__ == 'Void .ctor(Boolean)'
    assert DocWithCtorNoDocTest.TestMethod.__doc__ == 'Void TestMethod(Double, Int32)'
    assert DocWithCtorNoDocTest.StaticTestMethod.__doc__ == 'Void StaticTestMethod(Double, Int32)'


def test_doc_without_ctor():
    from Python.Test import DocWithoutCtorTest

    assert DocWithoutCtorTest.__doc__ == 'DocWithoutCtorTest Class'
    assert DocWithoutCtorTest.TestMethod.__doc__ == 'DocWithoutCtorTest TestMethod'
    assert DocWithoutCtorTest.StaticTestMethod.__doc__ == 'DocWithoutCtorTest StaticTestMethod'



import subprocess
import os
import platform

import pytest

from pythonnet.find_libpython import find_libpython
libpython = find_libpython()

pytestmark = pytest.mark.xfail(libpython is None, reason="Can't find suitable libpython")


def _run_test(testname):
    dirname = os.path.split(__file__)[0]
    exename = os.path.join(dirname, 'bin', 'Python.DomainReloadTests.exe')
    args = [exename, testname]

    if platform.system() != 'Windows':
        args = ['mono'] + args

    env = os.environ.copy()
    env["PYTHONNET_PYDLL"] = libpython

    proc = subprocess.Popen(args, env=env)
    proc.wait()

    assert proc.returncode == 0

def test_rename_class():
    _run_test('class_rename')

def test_rename_class_member_static_function():
    _run_test('static_member_rename')

def test_rename_class_member_function():
    _run_test('member_rename')

def test_rename_class_member_field():
    _run_test('field_rename')

def test_rename_class_member_property():
    _run_test('property_rename')

def test_rename_namespace():
    _run_test('namespace_rename')

def test_field_visibility_change():
    _run_test("field_visibility_change")

def test_method_visibility_change():
    _run_test("method_visibility_change")

def test_property_visibility_change():
    _run_test("property_visibility_change")

def test_class_visibility_change():
    _run_test("class_visibility_change")

def test_method_parameters_change():
    _run_test("method_parameters_change")

def test_method_return_type_change():
    _run_test("method_return_type_change")

def test_field_type_change():
    _run_test("field_type_change")

@pytest.mark.xfail(reason="Events not yet serializable")
def test_rename_event():
    _run_test('event_rename')

def test_construct_removed_class():
    _run_test("construct_removed_class")

def test_out_to_ref_param():
    _run_test("out_to_ref_param")

def test_ref_to_out_param():
    _run_test("ref_to_out_param")

def test_ref_to_in_param():
    _run_test("ref_to_in_param")

def test_in_to_ref_param():
    _run_test("in_to_ref_param")

def test_nested_type():
    _run_test("nested_type")

def test_import_after_reload():
    _run_test("import_after_reload")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# FIXME: FAIL: testImplicitAssemblyLoad AssertionError: 0 != 1

"""Basic stress test."""

from __future__ import print_function

import gc
import time
import unittest
# import pdb

try:
    import System
except ImportError:
    print("Load clr import hook")
    import clr
    clr.AddReference("Python.Test")
    clr.AddReference("System.Collections")
    clr.AddReference("System.Data")
    clr.AddReference("System.Management")


def main():
    start = time.clock()

    for i in range(2000):
        print(i)
        for name in (
            'test_module',
            'test_conversion',
            # 'test_class',
            'test_interface',
            'test_enum',
            'test_field',
            'test_property',
            'test_indexer',
            'test_event',
            'test_method',
            # 'test_delegate',
            'test_array',
        ):
            module = __import__(name)
            unittest.TextTestRunner().run(module.test_suite())

    # pdb.set_trace()

    stop = time.clock()
    took = str(stop - start)
    print('Total Time: {0}'.format(took))

    for i in gc.get_objects():
        print(i)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# FIXME: FAIL: testImplicitAssemblyLoad AssertionError: 0 != 1

"""Run all of the unit tests for this package over and over,
   in order to provide for better profiling.
"""

from __future__ import print_function

import gc
import os
import sys
import time

import runtests


def main():
    dirname = os.path.split(__file__)
    sys.path.append(dirname)

    gc.set_debug(gc.DEBUG_LEAK)

    start = time.clock()

    for i in range(50):
        print('iteration: {0:d}'.format(i))
        runtests.main()

    stop = time.clock()
    took = str(stop - start)
    print('Total Time: {0}'.format(took))

    for item in gc.get_objects():
        print(item, sys.getrefcount(item))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
TypeOffset is a C# class that mirrors the in-memory layout of heap
allocated Python objects.

This script parses the Python C headers and outputs the TypeOffset
C# class.

Requirements:
    - pycparser
    - clang
"""

from __future__ import print_function

import logging
import os
import sys
import sysconfig
import subprocess

if sys.version_info.major > 2:
    from io import StringIO
else:
    from StringIO import StringIO

from pycparser import c_ast, c_parser

_log = logging.getLogger()
logging.basicConfig(level=logging.DEBUG)

PY_MAJOR = sys.version_info[0]
PY_MINOR = sys.version_info[1]

# rename some members from their C name when generating the C#
_typeoffset_member_renames = {
    "ht_name": "name",
    "ht_qualname": "qualname"
}


def _check_output(*args, **kwargs):
    """Check output wrapper for py2/py3 compatibility"""
    output = subprocess.check_output(*args, **kwargs)
    if PY_MAJOR == 2:
        return output
    return output.decode("ascii")


class AstParser(object):
    """Walk an AST and determine the members of all structs"""

    def __init__(self):
        self._typedefs = {}
        self._typedecls = {}
        self._structs = {}
        self._struct_stack = []
        self._struct_members_stack = []
        self._ptr_decl_depth = 0
        self._struct_members = {}
        self._decl_names = {}

    def get_struct_members(self, name):
        """return a list of (name, type) of struct members"""
        defs = self._typedefs.get(name)
        if defs is None:
            return None
        node = self._get_leaf_node(defs)
        name = node.name
        if name is None:
            name = defs.declname
        return self._struct_members.get(name)

    def visit(self, node):
        if isinstance(node, c_ast.FileAST):
            self.visit_ast(node)
        elif isinstance(node, c_ast.Typedef):
            self.visit_typedef(node)
        elif isinstance(node, c_ast.TypeDecl):
            self.visit_typedecl(node)
        elif isinstance(node, c_ast.Struct):
            self.visit_struct(node)
        elif isinstance(node, c_ast.Decl):
            self.visit_decl(node)
        elif isinstance(node, c_ast.FuncDecl):
            self.visit_funcdecl(node)
        elif isinstance(node, c_ast.PtrDecl):
            self.visit_ptrdecl(node)
        elif isinstance(node, c_ast.IdentifierType):
            self.visit_identifier(node)

    def visit_ast(self, ast):
        for name, node in ast.children():
            self.visit(node)

    def visit_typedef(self, typedef):
        self._typedefs[typedef.name] = typedef.type
        self.visit(typedef.type)

    def visit_typedecl(self, typedecl):
        self._decl_names[typedecl.type] = typedecl.declname
        self.visit(typedecl.type)

    def visit_struct(self, struct):
        if struct.decls:
            self._structs[self._get_struct_name(struct)] = struct
            # recurse into the struct
            self._struct_stack.insert(0, struct)
            for decl in struct.decls:
                self._struct_members_stack.insert(0, decl.name)
                self.visit(decl)
                self._struct_members_stack.pop(0)
            self._struct_stack.pop(0)
        elif self._ptr_decl_depth:
            # the struct is empty, but add it as a member to the current
            # struct as the current member maybe a pointer to it.
            self._add_struct_member(struct.name)

    def visit_decl(self, decl):
        self.visit(decl.type)

    def visit_funcdecl(self, funcdecl):
        self.visit(funcdecl.type)

    def visit_ptrdecl(self, ptrdecl):
        self._ptr_decl_depth += 1
        self.visit(ptrdecl.type)
        self._ptr_decl_depth -= 1

    def visit_identifier(self, identifier):
        type_name = " ".join(identifier.names)
        self._add_struct_member(type_name)

    def _add_struct_member(self, type_name):
        if not (self._struct_stack and self._struct_members_stack):
            return

        # add member to current struct
        current_struct = self._struct_stack[0]
        member_name = self._struct_members_stack[0]
        struct_members = self._struct_members.setdefault(
            self._get_struct_name(current_struct), [])

        # get the node associated with this type
        node = None
        if type_name in self._typedefs:
            node = self._get_leaf_node(self._typedefs[type_name])
            # If the struct was only declared when the typedef was created, its member
            # information will not have been recorded and we have to look it up in the
            # structs
            if isinstance(node, c_ast.Struct) and node.decls is None:
                if node.name in self._structs:
                    node = self._structs[node.name]
        elif type_name in self._structs:
            node = self._structs[type_name]

        # If it's a struct (and not a pointer to a struct) expand
        # it into the current struct definition
        if not self._ptr_decl_depth and isinstance(node, c_ast.Struct):
            for decl in node.decls or []:
                self._struct_members_stack.insert(0, decl.name)
                self.visit(decl)
                self._struct_members_stack.pop(0)
        else:
            # otherwise add it as a single member
            struct_members.append((member_name, type_name))

    def _get_leaf_node(self, node):
        if isinstance(node, c_ast.Typedef):
            return self._get_leaf_node(node.type)
        if isinstance(node, c_ast.TypeDecl):
            return self._get_leaf_node(node.type)
        return node

    def _get_struct_name(self, node):
        return node.name or self._decl_names.get(node) or "_struct_%d" % id(node)


class Writer(object):

    def __init__(self):
        self._stream = StringIO()

    def append(self, indent=0, code=""):
        self._stream.write("%s%s\n" % (indent * "    ", code))

    def extend(self, s):
        self._stream.write(s)

    def to_string(self):
        return self._stream.getvalue()


def preprocess_python_headers():
    """Return Python.h pre-processed, ready for parsing.
    Requires clang.
    """
    fake_libc_include = os.path.join(os.path.dirname(__file__),
                                     "fake_libc_include")
    include_dirs = [fake_libc_include]

    include_py = sysconfig.get_config_var("INCLUDEPY")
    include_dirs.append(include_py)

    include_args = [c for p in include_dirs for c in ["-I", p]]

    defines = [
        "-D", "__attribute__(x)=",
        "-D", "__inline__=inline",
        "-D", "__asm__=;#pragma asm",
        "-D", "__int64=long long",
        "-D", "_POSIX_THREADS"
    ]

    if os.name == 'nt':
        defines.extend([
            "-D", "__inline=inline",
            "-D", "__ptr32=",
            "-D", "__ptr64=",
            "-D", "__declspec(x)=",
        ])

    if hasattr(sys, "abiflags"):
        if "d" in sys.abiflags:
            defines.extend(("-D", "PYTHON_WITH_PYDEBUG"))
        if "u" in sys.abiflags:
            defines.extend(("-D", "PYTHON_WITH_WIDE_UNICODE"))

    python_h = os.path.join(include_py, "Python.h")
    cmd = ["clang", "-pthread"] + include_args + defines + ["-E", python_h]

    # normalize as the parser doesn't like windows line endings.
    lines = []
    for line in _check_output(cmd).splitlines():
        if line.startswith("#"):
            line = line.replace("\\", "/")
        lines.append(line)
    return "\n".join(lines)



def gen_interop_head(writer):
    filename = os.path.basename(__file__)
    abi_flags = getattr(sys, "abiflags", "").replace("m", "")
    py_ver = "{0}.{1}".format(PY_MAJOR, PY_MINOR)
    class_definition = """
// Auto-generated by %s.
// DO NOT MODIFY BY HAND.

// Python %s: ABI flags: '%s'

// ReSharper disable InconsistentNaming
// ReSharper disable IdentifierTypo

using System;
using System.Diagnostics.CodeAnalysis;
using System.Runtime.InteropServices;

using Python.Runtime.Native;

namespace Python.Runtime
{""" % (filename, py_ver, abi_flags)
    writer.extend(class_definition)


def gen_interop_tail(writer):
    tail = """}
"""
    writer.extend(tail)


def gen_heap_type_members(parser, writer, type_name = None):
    """Generate the TypeOffset C# class"""
    members = parser.get_struct_members("PyHeapTypeObject")
    type_name = type_name or "TypeOffset{0}{1}".format(PY_MAJOR, PY_MINOR)
    class_definition = """
    [SuppressMessage("Style", "IDE1006:Naming Styles",
                     Justification = "Following CPython",
                     Scope = "type")]

    [StructLayout(LayoutKind.Sequential)]
    internal class {0} : GeneratedTypeOffsets, ITypeOffsets
    {{
        public {0}() {{ }}
        // Auto-generated from PyHeapTypeObject in Python.h
""".format(type_name)

    # All the members are sizeof(void*) so we don't need to do any
    # extra work to determine the size based on the type.
    for name, tpy in members:
        name = _typeoffset_member_renames.get(name, name)
        class_definition += "        public int %s  { get; private set; }\n" % name

    class_definition += """    }
"""
    writer.extend(class_definition)


def gen_structure_code(parser, writer, type_name, indent):
    members = parser.get_struct_members(type_name)
    if members is None:
        return False
    out = writer.append
    out(indent, "[StructLayout(LayoutKind.Sequential)]")
    out(indent, "internal struct %s" % type_name)
    out(indent, "{")
    for name, tpy in members:
        out(indent + 1, "public IntPtr %s;" % name)
    out(indent, "}")
    out()
    return True

def main():
    # preprocess Python.h and build the AST
    python_h = preprocess_python_headers()
    parser = c_parser.CParser()
    ast = parser.parse(python_h)

    # extract struct members from the AST
    ast_parser = AstParser()
    ast_parser.visit(ast)

    writer = Writer()
    # generate the C# code
    offsets_type_name = "NativeTypeOffset" if len(sys.argv) > 1 else None
    gen_interop_head(writer)

    gen_heap_type_members(ast_parser, writer, type_name = offsets_type_name)

    gen_interop_tail(writer)

    interop_cs = writer.to_string()
    if len(sys.argv) > 1:
        with open(sys.argv[1], "w") as fh:
            fh.write(interop_cs)
    else:
        print(interop_cs)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""Test if interop with recursive type inheritance works."""


def test_recursive_type_creation():
    """Test that a recursive types don't crash with a
    StackOverflowException"""
    from Python.Test import RecursiveInheritance

    test_instance = RecursiveInheritance.SubClass()
    test_instance.SomeMethod()
